
from typing import Union, Any, Optional
import pandas as pd
from evdspy.EVDSlocal.config.apikey_class import ApikeyClass
from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import default_start_date_fnc, default_end_date_fnc
from evdspy.EVDSlocal.index_requests.user_requests import ProxyManager, UrlBuilder, UrlSeries, ApiRequester, \
    DataProcessor, RequestConfig
def get_series_module():
    """get_series_module"""
    def initial_api_process_when_given(api_key: str = None) -> None:
        if api_key:
            if ApikeyClass().get_valid_api_key(check=False) is False:
                from evdspy.EVDSlocal.initial.load_commands_cmds_to_load import save_apikey
                save_apikey(api_key)
    def get_series(
            index: Union[str, tuple[Any, ...]],
            start_date: str = default_start_date_fnc(),
            end_date: str = default_end_date_fnc(),
            frequency: Union[str, int, None] = None,
            formulas: Union[str, tuple[str, int, ...]] = None,
            aggregation: Union[str, None] = None,
            cache: bool = False,
            proxy: Optional[str] = None,
            proxies: Optional[dict[str, str]] = None,
            debug: bool = False,
            api_key: Optional[str] = None
    ) -> Union[pd.DataFrame, RequestConfig]:
        """
        Retrieves economic data series from the specified API and returns it as a pandas DataFrame.
        Parameters
        ----------
        index : str or tuple of str
            The identifier(s) for the data series to fetch. Can be a single string for one series or a tuple of strings for multiple series.
        start_date : str, optional
            The start date for the data retrieval in 'DD-MM-YYYY' format, by default calls default_start_date_fnc().
        end_date : str, optional
            The end date for the data retrieval in 'DD-MM-YYYY' format, by default calls default_end_date_fnc().
        frequency : str, optional
            The frequency at which data should be retrieved.
            monthly | weekly | annually | semimonthly | semiannually | business
        formulas : str or tuple of str, optional
            The computation methods to apply to the data series
            level | percentage_change | difference | year_to_year_percent_change | year_to_year_differences
        aggregation : str or tuple of str, optional
            The aggregation methods to apply to the data, similar to formulas.
            avg |min | max | first | last | sum
        cache : bool, optional
            If True, uses cached data when available to speed up the data retrieval process, by default False.
        proxy : str, optional
            The URL of the proxy server to use for the requests, by default None.
        proxies : dict, optional
            A dictionary of proxies to use for the request, by default None.
        debug : bool, optional
            If True, runs the function in debug mode, providing additional debug information without making a real API request, by default False.
        api_key : str, optional
            The API key required for accessing the data, by default None.
            When it was given for the first time it will be saved to a file for the subsequent requests.
            alternatively it may be saved by save("APIKEY") function or $ evdspy save [from console]
        Returns
        -------
        pd.DataFrame
            A pandas DataFrame containing the retrieved data series.
        Raises
        ------
        ValueError
            If an invalid API key is provided or required parameters are missing.
        Examples
        --------
        Basic usage:
        >>> template = "TP.ODEMGZS.BDTTOPLAM"
        >>> df = get_series(template, start_date="01-01-2020", end_date="01-01-2021", frequency="monthly")
        >>> print(df.head())
        Using multiple indexes and cache:
        >>> indexes = ("TP.ODEMGZS.BDTTOPLAM", "TP.ODEMGZS.ABD")
        >>> df = get_series(indexes, start_date="01-01-2020", frequency="monthly", cache=True)
        >>> print(df.head())
        Applying formulas and aggregation:
        >>> df = get_series(template, start_date="01-01-2020", formulas="level", aggregation="sum")
        >>> print(df.head())
        """
        """initial_api_process_when_given"""
        initial_api_process_when_given(api_key)
        """RequestConfig"""
        config = RequestConfig(index=index,
                               start_date=start_date,
                               end_date=end_date,
                               frequency=frequency,
                               formulas=formulas,
                               aggregation=aggregation,
                               cache=cache
                               )
        """ProxyManager"""
        proxy_manager = ProxyManager(proxy=proxy, proxies=proxies)
        """UrlBuilder"""
        url_builder = UrlBuilder(config, url_type=UrlSeries())
        """ApiRequester"""
        api_requester = ApiRequester(url_builder, proxy_manager)
        if debug:
            return api_requester.dry_request()
        """DataProcessor"""
        data_processor = DataProcessor(api_requester())
        return data_processor()
    def test_get_series2(capsys):
        with capsys.disabled():
            # setup()
            df = get_series("TP.ODEMGZS.BDTTOPLAM",
                            cache=False)
            assert isinstance(df, pd.DataFrame)
    def t_stream():
        import streamlit as st
        df = get_series("TP.ODEMGZS.BDTTOPLAM",
                        cache=True)
        st.write(df)
__all__ = (
    'get_series_module',
)


import pandas as pd

from evdspy import get_series_exp
from evdspy.EVDSlocal.index_requests.get_series_indexes_exp import Result

index1 = "TP.GSYIH02.GY.CF"
index2 = "TP.GSYIH02.GY.CF"

indexes = """TP_GSYIH01_GY_CF
TP_GSYIH02_GY_CF
TP_GSYIH03_GY_CF
TP_GSYIH04_GY_CF
TP_GSYIH05_GY_CF
TP_GSYIH06_GY_CF
TP_GSYIH07_GY_CF
TP_GSYIH08_GY_CF
TP_GSYIH09_GY_CF
TP_GSYIH10_GY_CF
TP_GSYIH11_GY_CF
TP_GSYIH14_GY_CF
TP_GSYIH15_GY_CF
TP_GSYIH16_GY_CF
"""
index_table = "bie_gsyhgycf"

from typing import Any
import time


def fnc(item: Any) -> Result:
    print(f"  Checking ...  {item}")
    time.sleep(2)
    df = get_series_exp(item, debug=False, cache=True)
    print(df)
    return df


DF = pd.DataFrame


def test_get_series_exp(capsys):
    with capsys.disabled():
        items = [index1, index2, indexes, index_table]

        for item in items:
            result: Result = fnc(item)
            # we will check if this function handles all data types correctly
            # then examine the last one to see what we get as a result from the function
            assert isinstance(result.data, DF)
            assert isinstance(result.metadata, DF)
            assert callable(result.write)
            assert callable(result.to_excel)




from ..messages.error_classes import *
from evdspy.EVDSlocal.helper.print_m import *
from evdspy.EVDSlocal.initial_setup.setup_folders import check_folders_setup_necessary
if not check_folders_setup_necessary():
    print("setup checked...")
    try:
        from .load_commands_ import *
        # help_evds()
        # setup_now()
        # help_evds()
    except OptionsFileNotLoadedError:
        print("load_commands_ error trying setup")
        time.sleep(3)
        from ..initial_setup.initial_setups import *
        SetupInitial().setup()
else:
    from .load_commands_ import *
    # help_evds()


# ------------------------------------------------------
#
#       __init__
#                           package: evdspy @2022
# ------------------------------------------------------
from evdspy.EVDSlocal.main_ import *
from evdspy.EVDSlocal.index_requests.get_series_indexes import (

    get_series
)

from evdspy.EVDSlocal.index_requests.get_series_indexes_exp import get_series_exp

from evdspy.EVDSlocal.utils.utils_general import ls

from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import (
    default_start_date_fnc,
    default_end_date_fnc,
    correct_types,
)

from evdspy.EVDSlocal.common.clear_cache_folder import clear_cache


# __all__ = [
#         "default_start_date_fnc",
#         "default_end_date_fnc",
#         "correct_types",
#         "get_series" ,
#         "ls"
#     'easy_setup',
#     'setup',
#     'setup_series',
#     "setup_series_steps",
#     'help_evds',
#     'check',
#     'get',
#     'h',
#     'help_evdspy',
#     'help_',
#     'create_series_file',
#     'csf',
#     'show',
#     'get_df_test',
#     'console',
#     'menu',
#     'menu_onload',
#     'version',
#     'save_apikey',
#     'save',
#     'remove_cache',
#     'create_options_file',
#     'cof',
#     'console_main',
#     'get_df_datagroup',
#     'get_datagroups_with_code',
#     'get_category_name' ,
#     'get_series_list_of_subject'
# ]




from evdspy.EVDSlocal.main_ import *
def test_load_general():
    assert callable(setup_series), "setup_series is not callable"
    assert callable(help_), "help_ is not callable"
def test_setup_now():
    # evdspy.setup_now()
    # evdspy.help_()
    # evdspy.check()
    easy_setup()
    get()
from evdspy.EVDSlocal.config.apikey_class import ApikeyClass
def test_set_apikey():
    assert ApikeyClass().set_api_key_filetype(value="test") == "test", "set api key did not work"
    assert ApikeyClass().set_api_key_runtime(value="testruntime") == "testruntime", "set api key did not work"
def test_get_apikey():
    assert ApikeyClass().set_api_key_filetype(value="test") == "test", "get api key did not work"
    assert ApikeyClass().get_api_key_fromfile() == "test", "get api key did not work"
    assert ApikeyClass().set_api_key_filetype("test3") == "test3", "set api key did not work"
    assert ApikeyClass().get_api_key_fromfile() == "test3", "get api3 key did not work"
    assert ApikeyClass().get_api_key_runtime() == "testruntime", "get testruntime  key did not work"
def test_valid_key():
    assert ApikeyClass().set_api_key_filetype(value="thatisfromFile") == "thatisfromFile", "get api key did not work"
    ApikeyClass().now_testing_is_key_is_valid = "thisisValid"
    # assert ApikeyClass().get_valid_api_key() == "thisisValid", "get thisisValid  key did not work"



# ------------------------------------------------------
#
#       main_.py
#                           package: evdspy @2022
# ------------------------------------------------------
from .initial.start_args import *
import typing as t
def print_with_failure_style(msg):
    print(msg)
def print_with_success_style(msg):
    print(msg)
def check_compat():
    import platform
    v_tuple = platform.python_version_tuple()
    # v_tuple = "3.6.0".split(".")
    v_tuple = tuple(map(lambda x: int(x), v_tuple))
    v = sys.version  # sys.version_info
    if (4, 1, -1) < v_tuple:
        print_with_failure_style(
            f"Your python version is {v}. This program may break because it is currently only compatible with versions between 3.7 and 3.11")
        return False
    elif (3, 7, 0) > v_tuple:
        print_with_failure_style(
            f"Your python version is {v}. This program may break because it is currently only compatible with versions between 3.7 and 3.11")
        return False
    else:
        # print_with_success_style(
        #     f"Your python version is {v} This program was tested with this version and runs properly. However, "
        #     f"if you notice a bug or if your version breaks at runtime please feel free to open a PR on github.")
        return True
if not check_compat():
    import sys
    sys.exit(0)
def register_actions(actions=None) -> tuple:
    if actions is None:
        actions = []
    """register_actions"""
    from evdspy.EVDSlocal.initial_setup.initial_setups import SetupInitial
    actions.append(SetupInitial().setup)
    # actions.append( SetupInitial().create_series_text_ex   )
    return tuple(actions)
def do_start(actions: t.Tuple[t.Callable]):
    """CHECK => SETUP , START ==> or => START"""
    from evdspy.EVDSlocal.initial_setup.setup_folders import check_folders_setup_necessary
    if not check_folders_setup_necessary():
        for action in actions:
            if callable(action):
                
                action()
def initial_checks_after_first_run():
    """initial_checks_after_first_run"""
    assert callable(check)
    assert callable(get)
    assert callable(menu)
# --------------------------------- M A I N ---------
# do_start(register_actions())
from evdspy.EVDSlocal.initial.load_commands_ import *
initial_checks_after_first_run()
# --------------------------------- / M A I N ---------



current_menu_appears = """
        Quickly setup your work environment run commands from the menu.
        Currently you may choose one below or create your own functions and
        add to this list following instructions on Readme.md
        --------------------------------------------------
                             M E N U
        --------------------------------------------------
                       1. check setup
                       2. setup
                       3. create user options file
                       4. create series file
                       5. add new series group
                       6. get data
                       7. get categories
                       8. help
                       9. show api key
                       10. save api key to file
                       11. remove cache folders
                       12. evdspy as a command line prompt
                       13. version
                       14. console
"""



from inspect import getframeinfo, stack
import builtins as __builtin__
from evdspy.EVDSlocal.initial.start_options import *
if DEBUG_PRINT :
    """
        Here we will overload print
        see inital.start_options.py
    """
    def debugInfo(s="", *args, **kw):
        caller = getframeinfo(stack()[2][0])
        msg = "---called from : {} - line number :  {} -- {} ".format(caller.filename, caller.lineno, s)
        return msg
    def print(*args, **kwargs):
        if DEBUG_PRINT:
            d = debugInfo(*args)
            __builtin__.print("XXX", d)
            return __builtin__.print(*args, **kwargs)
        return __builtin__.print(*args, **kwargs)


import os
import shutil
from pathlib import Path  

def clear_cache_folder(cache_folder_path : Path ):
    if 'cache/evdspy' not in str(cache_folder_path) :
        
        print('This does not look like cache folder. I can only delete content of cache folder of evdspy package')
        return 
    if not os.path.exists(cache_folder_path):
        print(f"The folder {cache_folder_path} does not exist.")
        return

    for filename in os.listdir(cache_folder_path):
        file_path = os.path.join(cache_folder_path, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print(f"Failed to delete {file_path}. Reason: {e}")

    print(f"All files and subdirectories in [{cache_folder_path}] have been deleted.")


def clear_cache():
    from pathlib import Path   
    cache_folder = Path.home() / ".cache" / "evdspy"
    clear_cache_folder(cache_folder)




from ..common.common_imports import *
from dataclasses import dataclass, field
from ..utils.utils_general import *
from ..components.options_class import Options
from evdspy.EVDSlocal.initial_setup.api_key_save import get_api_key_from_file_improved
from evdspy.EVDSlocal.config.apikey_class import *
from ..components.options_class import load_options
from typing import Optional 


def get_if_attr_exist(pred, f):
    if pred:
        return f(pred)
    return False
@dataclass
class Credentials:
    options_: Options = field(default_factory=load_options)
    apiKey: Optional[any] = False
    proxy = False
    address: str = "https://evds2.tcmb.gov.tr/service/evds/"
    def __post_init__(self):
        self.apikeyobj = field(default_factory=ApikeyClass)  # ApikeyClass()
        self.get_api_keys()
        if hasattr(self.options_, "proxy_file"):
            self.proxy = get_proxy_from_file(self.options_.proxy_file)
    def get_api_keys(self):
        self.apikeyobj = ApikeyClass()
        # from file
        api_from_file = get_api_key_from_file_improved()
        if api_from_file:
            self.apikeyobj.set_api_key_filetype(api_from_file)
    def __repr__(self):
        return f"""
        apikey: obscured(read_successfully)
        proxy : obscured(read_successfully)
        adres:{self.address}
        """
__all__ = [
    'Credentials'
]


from pathlib import Path
import pandas as pd
from evdspy.EVDSlocal.index_requests.get_series_indexes import (
    default_start_date_fnc,
    default_end_date_fnc,
    get_series,
)
from evdspy.EVDSlocal.utils.utils_general import get_env_api_key
from evdspy.EVDSlocal.utils.utils_test import (
    ApiClassWhileTesting,
    get_api_key_file,
    skip_if_gthub,
    skip_if_not_keyvalid,
    is_df,
)
from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import (
    Frequency,
    freq_enum,
    Formulas,
    AggregationType,
    correct_types,
)
from evdspy.EVDSlocal.utils.utils_test import get_api_key_while_testing

# from evdspy.EVDSlocal.index_requests.user_requests import
from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import (
    default_start_date_fnc,
    default_end_date_fnc,
)
from evdspy.EVDSlocal.index_requests.user_requests import RequestConfig

import os

# from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import *

try:
    import pytest
except ImportError:
    pass


@skip_if_gthub
def test_get_apikey_while_testing():

    a = get_api_key_while_testing()
    assert a


@skip_if_gthub
def test_get_api_key_file(capsys):
    with capsys.disabled():
        api_key_file = get_api_key_file(deep=8)
        print(api_key_file.absolute())
        assert api_key_file is not None


@skip_if_gthub
def test_ApiClassWhileTesting(capsys):
    with capsys.disabled():
        api_key = ApiClassWhileTesting().key
        # print(ApikeyClass().obscure(api_key))
        print(Path("").absolute())


@skip_if_not_keyvalid
@skip_if_gthub
def test_get_series_bridge(capsys):
    with capsys.disabled():
        df = get_series(
            "bie_gsyhgycf", cache=False, api_key=get_env_api_key(check=True)
        )
        assert is_df(df)


@skip_if_not_keyvalid
def test_get_diff_series(capsys):
    with capsys.disabled():
        template = """TP_GSYIH01_GY_CF
        TP_GSYIH02_GY_CF
        TP_GSYIH03_GY_CF
        TP_GSYIH04_GY_CF
        TP_GSYIH05_GY_CF
        TP_GSYIH06_GY_CF
        TP_GSYIH07_GY_CF
        TP_GSYIH08_GY_CF
        TP_GSYIH09_GY_CF
        TP_GSYIH10_GY_CF
        TP_GSYIH11_GY_CF
        TP_GSYIH14_GY_CF
        TP_GSYIH15_GY_CF
        TP_GSYIH16_GY_CF
        """
        df = get_series(
            template, debug=False, cache=False, api_key=get_env_api_key(check=True)
        )
        assert is_df(df)


# @skip_if_not_keyvalid
def test_template_series(capsys):
    with capsys.disabled():
        balance_of_pay1 = "TP.ODEMGZS.BDTTOPLAM", "TP.ODEMGZS.ABD"
        balance_of_pay2 = """
        TP.ODEMGZS.BDTTOPLAM #
        TP.ODEMGZS.ABD #
        """
        a1 = get_series(balance_of_pay1, debug=True)
        a2 = get_series(balance_of_pay2, debug=True)
        print(a1.hash, a2.hash)
        assert a1 == a2


@skip_if_not_keyvalid
def test_a(capsys):
    with capsys.disabled():
        balance_of_pay1 = "TP.ODEMGZS.BDTTOPLAM", "TP.ODEMGZS.ABD"
        balance_of_pay2 = """
        TP.ODEMGZS.BDTTOPLAM #
        TP.ODEMGZS.ABD #
        """
        a1 = get_series(balance_of_pay1, debug=True)
        a2 = get_series(balance_of_pay2, debug=True)
        assert a1 == a2
        print(a1)

        a1 = get_series(balance_of_pay1, debug=False)
        print(a1)
        assert is_df(a1)


def test_template_series2(capsys):
    with capsys.disabled():
        balance_of_pay1 = "TP.ODEMGZS.BDTTOPLAM", "TP.ODEMGZS.ABD"
        balance_of_pay2 = """
        TP.ODEMGZS.BDTTOPLAM #
        TP.ODEMGZS.ABD #
        """
        a1 = get_series(balance_of_pay1, aggregation="avg", debug=True)
        a2 = get_series(balance_of_pay2, aggregation=("avg", "avg"), debug=True)
        print(a1.hash, a2.hash)
        assert a1 == a2


# @skip_if_not_keyvalid


def test_template_series3(capsys):
    with capsys.disabled():
        balance_of_pay1 = "TP.ODEMGZS.BDTTOPLAM", "TP.ODEMGZS.ABD"
        balance_of_pay2 = """
        TP.ODEMGZS.BDTTOPLAM #
        TP.ODEMGZS.ABD #
        """
        a1 = get_series(balance_of_pay1, formulas="level", debug=True)
        a2 = get_series(balance_of_pay2, formulas=(0, 0), debug=True)
        print(a1.hash, a2.hash)
        assert a1 == a2


def test_freq(capsys):
    f = Frequency.annually
    with capsys.disabled():
        assert f() == "&frequency=8"
        print("\n")
        print(f())
        g = freq_enum("daily")
        assert g == "&frequency=1"
        assert freq_enum("daily") == "&frequency=1"
        assert freq_enum("monthly") == "&frequency=5"
        assert freq_enum("annual") == "&frequency=8"
        assert freq_enum(3) == "&frequency=3"
        assert freq_enum(1) == "&frequency=1"


def test_pickles():
    pickle_folder = Path.home() / ".cache" / "evdspy"
    os.makedirs(pickle_folder, exist_ok=True)


# def is_testing():
#     return GithubActions().is_testing() and not key_valid()


@skip_if_not_keyvalid
def test_get_series2(capsys):
    with capsys.disabled():
        df = get_series("TP.ODEMGZS.BDTTOPLAM", cache=False)
        assert isinstance(df, pd.DataFrame)


# def test_get_df_datagroup(capsys):
#     # if not GithubActions().is_testing():
#     #     return
#     from evdspy.main import get_df_datagroup
#     with capsys.disabled():
#         df = get_df_datagroup(
#             datagroup="bie_gsyhgycf",
#             start_date="01-01-1998",
#             end_date="01-01-2030",
#         )
#         print(df)


# def test_get_api_key_while_testing(capsys):
#     with capsys.disabled():
#         a = ApiClassWhileTesting().key
#         assert len(a) > 5 and "lg" in a


@skip_if_not_keyvalid
def test_get_series(capsys):
    with capsys.disabled():
        index = "TP.YSSK.A1"
        start_date = "01-01-2000"
        end_date = "01-01-2100"
        cache = True
        df = get_series(
            index, start_date, end_date, cache, api_key=ApiClassWhileTesting()()
        )
        print(df)
        assert is_df(df)


# from evdspy.EVDSlocal.index_requests.get_series_indexes import Formulas, correct_types, AggregationType
# @skip_if_not_keyvalid
def test_aggr_types(capsys):
    balance_of_pay1 = "TP.ODEMGZS.BDTTOPLAM", "TP.ODEMGZS.ABD"
    balance_of_pay2 = """
    TP.ODEMGZS.BDTTOPLAM #
    TP.ODEMGZS.ABD #
    """
    cache = True
    with capsys.disabled():
        u1 = RequestConfig(
            balance_of_pay1,
            frequency="weekly",
            start_date=default_start_date_fnc(),
            end_date=default_end_date_fnc(),
            aggregation=("avg",),
            # proxy="http://127.0.0.1:8000",
            # proxies={"http": "http://127.0.0.1:8000"},
            cache=cache,
        )
        u2 = RequestConfig(
            balance_of_pay2,
            frequency="weekly",
            start_date=default_start_date_fnc(),
            end_date=default_end_date_fnc(),
            aggregation=("avg",),
            # proxy="http://127.0.0.1:8000",
            # proxies={"http": "http://127.0.0.1:8000"},
            cache=cache,
        )
        assert u1 == u2


def test_correct(capsys):
    with capsys.disabled():
        assert hasattr(Formulas, "from_str")
        assert Formulas.level.value == 0
        assert Formulas.from_str("level").value == 0
        assert correct_types("level", Formulas) == 0
        assert correct_types(
            (
                "level",
                "level",
            ),
            enum_class=Formulas,
        ) == (
            0,
            0,
        )


def test_correct2(capsys):
    with capsys.disabled():
        assert correct_types("avg", AggregationType) == "avg"
        assert correct_types(
            (
                "avg",
                "min",
            ),
            AggregationType,
        ) == (
            "avg",
            "min",
        )


@skip_if_not_keyvalid
def test_mixedcase_get_series(capsys):
    index = """
    tp.sekbil1122.GENEL
    tp.sekbil1122.A
    tp.sekbil1122.01
    tp.sekbil1122.02
    TP.SEKbil1122.03
    TP.sekBIL1122.B
    TP.SEKBIL1122.05
    TP.SEKBIL1122.07
    TP.sekbIL1122.08
    TP.SEKBIL1122.09
    TP.SEKBIL1122.C
    """
    with capsys.disabled():
        df = get_series(index)
        assert is_df(df)


@skip_if_not_keyvalid
def test_gets_upper(capsys):
    with capsys.disabled():
        template = """TP_GSYIH01_GY_CF
                tp_gsyih02_gy_cf
                tp_gsyih03_gy_cf
                tp_gsyih04_gy_cf
                tp_gsyih05_GY_CF
                TP_GSYIH06_GY_CF
                TP_GSYIH07_GY_CF
                TP_GSYIH08_GY_CF
                TP_GSYIH09_GY_CF
                TP_GSYIH10_GY_CF
                TP_GSYIH11_GY_CF
                TP_GSYIH14_GY_CF
                TP_GSYIH15_GY_CF
                TP_GSYIH16_GY_CF
        """
        df = get_series(template)
        assert is_df(df)


# @skip_if_not_keyvalid
# def test_multi(capsys):
#     template = """
# bie_sekbil1122
# bie_sekbil1001
# bie_sekbil3111
# bie_sekbil3001
# bie_sekbil3051
#     """
#     names = tuple(x for x in template.splitlines() if x.strip())
#     print(names)
#     dfs = tuple(map(get_series, names))
#     assert all(map(lambda x: is_df(x), dfs))
#     tuple(map(lambda x: print(x.shape), dfs))


@skip_if_not_keyvalid
def test_get_series_b(capsys):
    index = """
    TP.OSUVBG01 # some comments
TP.OSUVBG02 # ...
TP.OSUVBG03
    TP.OSUVBG04
    TP.OSUVBG05
    TP.OSUVBG06 #
    TP.OSUVBG07  # some comments
    TP.OSUVBG08 #
    TP.OSUVBG09   # some more comments
 TP.OSUVBG10
TP.OSUVBG11
TP.OSUVBG12
TP.OSUVBG13
TP.OSUVBG14
    """
    with capsys.disabled():
        df = get_series(index)
        assert is_df(df)




from .load_commands_cmds_to_load import get_df_datagroup
from evdspy import *
df_exchange_rate = get_df_datagroup(
    datagroup="bie_dkdovytl",
    start_date='01-01-2010',
    end_date='01-01-2030',
)



# ------------------------------------------------------------------------------
import requests as requests
from evdspy.EVDSlocal.utils.utils_general import *
from evdspy.EVDSlocal.requests_.my_cache import MyCache
from evdspy.EVDSlocal.messages.error_classes import *
from evdspy.EVDSlocal.config.apikey_class import *
from evdspy.EVDSlocal.config.config import config
from evdspy.EVDSlocal.stats.save_stats_of_requests import *
# ------------------------------------------------------------------------------
m_cache = MyCache()
CANCEL_REQUEST_TEMP = config.cancel_request_temp
from evdspy.EVDSlocal.requests_.mock_req import *
class RealRequest(object):
    """ pytest will get this mock response object"""
    def __init__(self, url: str, proxies=None, api_key=False):
        self.url = url
        self.proxies = proxies
        if not api_key:
            self.api_key = ApikeyClass().key
        else:
            self.api_key = api_key
    def request_no_proxy(self) -> requests.Response:
        return requests.get(self.url)
    def request_proxy(self) -> requests.Response:
        return requests.get(self.url, proxies=self.proxies)
    def request(self) -> requests.Response:
        if config.current_mode_is_test and not config.temp_cancel_mock_request:
            """ pytest will get this mock response object"""
            return mock_request(self.url, self.proxies)
        if self.proxies is None:
            return self.request_no_proxy()
        return self.request_proxy()
class RealRequestWithParam(RealRequest):
    def get_header(self) -> dict:
        # api_key = ApikeyClass().key
        headers = {
            "key": self.api_key
        }
        return headers
    def request_no_proxy(self) -> requests.Response:
        headers = self.get_header()
        # print(headers)
        # print(self.url)
        response = requests.get(self.url, headers=headers)
        return response
    def request_proxy(self) -> requests.Response:
        headers = self.get_header()
        return requests.get(self.url, headers=headers, proxies=self.proxies)
    def request(self) -> requests.Response:
        if config.current_mode_is_test and not config.temp_cancel_mock_request:
            """ pytest will get this mock response object"""
            return mock_request(self.url, self.proxies)
        if self.proxies is None:
            return self.request_no_proxy()
        return self.request_proxy()


from evdspy.EVDSlocal.common.files import Read
from pathlib import Path

from typing import Union

"""
Usage :
    SingletonOptions().set("default_start_date_user", "UserStartDate")
    SingletonOptions().set("default_end_date_user", "UserEndDate")
    SingletonOptions().set("default_cache_user", "UserCache")
    SingletonOptions().get_valid_value("default_cache")
    SingletonOptions().get_valid_value("default_start_date")
    SingletonOptions().get_valid_value("default_end_date")
    SingletonOptions().get_valid_value("Avoid_Absolute_Paths")
"""
if "__main__" == __name__:
    if True:
        default_series_file_name = "config_series.cfg"
        default_data_folder_name = "SeriesData"
        Avoid_Absolute_Paths_ = True
        Default_Prefix_ = "EVPY_"
        # options_folder_name = r"IO"
        DEBUG_LOG_CANCEL = False
        DEGUB_NOTICE = True
        DEBUG_PRINT = False
        default_cache = "daily"  # nocache / hourly
        default_end_date = "01-12-2030"
        default_start_date = "01-01-2019"
else:
    from ..config.config import *
    from ..initial.start_options import *
class Options:
    def __init__(self,
                 default_start_date: str,
                 default_start_date_user: Union[str, None],
                 default_end_date: str,
                 default_end_date_user,
                 default_cache: str,
                 default_cache_user: str,
                 avoid_absolute_paths: bool = True,
                 avoid_absolute_paths_user: bool = True,
                 ):
        self.default_start_date = default_start_date
        self.default_start_date_user = default_start_date_user
        self.default_end_date = default_end_date
        self.default_end_date_user = default_end_date_user
        self.default_cache = default_cache
        self.default_cache_user = default_cache_user
        self.avoid_absolute_paths = avoid_absolute_paths
        self.avoid_absolute_paths_user = avoid_absolute_paths_user

def load_options():
    options_ = Options(
        default_start_date=default_start_date,
        default_start_date_user=None,
        default_end_date=default_end_date,
        default_end_date_user=None,
        default_cache=default_cache,
        default_cache_user=None,
        avoid_absolute_paths=True,
        avoid_absolute_paths_user=True
    )
    return options_
class SingletonOptions(object):
    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(SingletonOptions, cls).__new__(cls)
            cls.post_init(cls)
        return cls.instance
    def post_init(cls):
        cls.options_ = load_options()
        # read_user_options_on_load()
    def check(cls):
        disp_message = f"""
       default_cache            :  {cls.options_.default_cache}
       default_cache (User)     :  {cls.options_.default_cache_user}
       default start date       :  {cls.options_.default_start_date}
       default start (User)     :  {cls.options_.default_start_date_user}
       default end date         :  {cls.options_.default_end_date}
       default end date (User)  :  {cls.options_.default_end_date_user}
       Avoid Absolute Paths         :  {cls.options_.avoid_absolute_paths}
       Avoid Absolute Paths (User)  :  {cls.options_.avoid_absolute_paths_user}
"""
        return disp_message
    def getitem(cls, attr, default_value=None):
        value = getattr(cls.options_, attr, default_value)
        return value
    def set(cls, attr, value):
        setattr(cls.options_, attr, value)
    def __str__(self):
        # print(self.options_.__dict__)
        return str(self.options_.__dict__)
    def display(self):
        return str(self)
        # cls.options_.(attr) = cls.options_.metadata.( attr)
    def get_valid_value(self, attr: str):
        """ """
        user_preferred_value_or_default: str = self.getitem(f"{attr}_user", None)
        if not user_preferred_value_or_default:
            user_preferred_value_or_default = self.getitem(f"{attr}", None)
        return user_preferred_value_or_default
notsetyet = "not set yet OptionsClassfile"
class NotSetYet:
    def __str__(self):
        return "not set yet"
    def __eq__(self, other):
        return False
    def __and__(self, other):
        return False
    def __bool__(self):
        return False
def read_user_options_on_load():
    """below this file it will be called"""
    opts = SingletonOptions()
    file_name = "options.cfg"
    if not Path(file_name).is_file():
        # print(Path(file_name))
        cls = SingletonOptions()
        SingletonOptions().set("default_cache_user", cls.options_.default_cache)
        SingletonOptions().set("default_start_date_user", cls.options_.default_start_date)
        SingletonOptions().set("default_end_date_user", cls.options_.default_end_date)
        SingletonOptions().set("avoid_absolute_paths_user", cls.options_.avoid_absolute_paths_user)
        # raise "file"
        # print("user options file was not found. Program will continue with default options...")
        return
    def read_options_file():
        cont = Read(file_name, "could not read ")
        lines = [x for x in cont.splitlines() if not x.strip().startswith("#")]
        lines = [x.split("#")[0] for x in lines]
        return tuple(lines)
    def get_item_from_content(lines: tuple, item: str = None):
        if item is None:
            return None
        for line in lines:
            if item in line and ":" in line and line.split(":")[0].strip() == item:
                return line.split(":")[1].strip()
        return NotSetYet()
    lines = read_options_file()
    default_cache_user = get_item_from_content(lines, "cache_freq")
    default_start_date_user = get_item_from_content(lines, "gl_date_start")
    default_end_date_user = get_item_from_content(lines, "gl_date_end")
    default_Avoid_Absolute_Paths_user_string = str(get_item_from_content(lines, "avoid_absolute_paths"))
    default_avoid_absolute_paths_user = NotSetYet()
    # default_avoid_absolute_paths_user = NotSetYet() #  'Not set yet'
    if default_Avoid_Absolute_Paths_user_string.strip().lower() == "false":
        default_avoid_absolute_paths_user = False
    if default_Avoid_Absolute_Paths_user_string.strip().lower() == "true":
        default_avoid_absolute_paths_user = True
    SingletonOptions().set("default_cache_user", default_cache_user)
    SingletonOptions().set("default_start_date_user", default_start_date_user)
    SingletonOptions().set("default_end_date_user", default_end_date_user)
    SingletonOptions().set("avoid_absolute_paths_user", default_avoid_absolute_paths_user)
def notest():
    a1 = SingletonOptions()
    print(a1.display())
    a2 = SingletonOptions()
    a2.options_.default_cache = "smt else "
    print(a1.display())
    print(a2.display())
    a3 = SingletonOptions()
    print(a3.display())
    a3.set("default_start_date", "1212asd54")
    print(a3.display())
    print(a3.get_valid_value("default_start_date"))
    print(a3.get_valid_value("default_cache"))
# notest()
read_user_options_on_load()




from evdspy.EVDSlocal.utils.utils_general import *
from evdspy.EVDSlocal.series_format.series_format_config_funcs import *
mainSepBegin = "---Series---------------------------------"
mainSepEnd = "---/Series---------------------------------"
GSEP = "--++--"
NEW_LINE = "\n"
# "cache_freq gl_date_start gl_date_end"
# ----------------------------- C O N F I G  ---------------------------------------------------
items_from_user_config = [
    "Cache frequency",
    "global date start for series",
    "global date end for the series",
    "avoid absolute paths protection" ,
]
explanations_config = [
    f"Cache frequency will be used to decide to make a new request. {NEW_LINE} "
    f"If recent cache is found in local cache folder program will use it instead of making a request. {NEW_LINE}"
    f" {indent} Options : monthly / daily / nocache. Default : daily {NEW_LINE}",
    f"if not special start date for the series was given in `config_series.cfg` file "
    f"this date will be used as start date of series {NEW_LINE} Example 19-01-2015",
    f"if not special end date for the series was given in `config_series.cfg` file "
    f"this date will be used as end date of series {NEW_LINE} Example : 19-01-2030",
    r"if set False it will accept absolute paths such as C:\Users\Userx . default is True to protect users other folders from polluting"
]
check_funcs_options = [TrueFunc, TrueFunc, TrueFunc, TrueFunc]
default_answers_config = ['', '', '', 'True']
same = lambda x: x
transform_answers_options = (same for _ in items_from_user_config)


# typings .......................................................... typings
from typing import Union

import pandas as pd

T_str_int_None = Union[str, int, None]
T_str_tuple_None = Union[str, tuple[str], None]
T_tuple_str_int_None = Union[str, int, tuple[str], tuple[int], None]
T_maybeDf = Union[pd.DataFrame, bool, None]





from dataclasses import dataclass
from abc import ABC
from enum import Enum
from typing import Callable, List, Tuple, Union
from datetime import datetime
from evdspy.EVDSlocal.components.api_params_enums import *
default_api_domain = "https://evds2.tcmb.gov.tr/service/evds/"
@dataclass
class ApiDomain(ABC):
    domain: str
@dataclass
class EVDSApiDomain(ApiDomain):
    domain: str = default_api_domain
    def __repr__(self):
        return self.domain
@dataclass
class EVDSApiDomainCategories(ApiDomain):
    domain: str = f"{default_api_domain}categories/"
    def __repr__(self):
        return self.domain
@dataclass
class EVDSApiDomainDatagroups(ApiDomain):
    domain: str = f"{default_api_domain}datagroups/"
    def __repr__(self):
        return self.domain
@dataclass
class EVDSApiDomainDatagroupIndividual(ApiDomain):
    domain: str = f"{default_api_domain}"  # datagroup=bie_yssk
    def __repr__(self):
        return self.domain
# GeneralIndexesDatagroupSeriesList
@dataclass
class EVDSApiDominDatagroupSeriesList(ApiDomain):
    domain: str = f"{default_api_domain}serieList/"  # serieList/key=xxx&type=json&code=bieyssk
    def __repr__(self):
        return self.domain
@dataclass
class UrlParam(ABC):
    """UrlParam class """
    value: Union[str, datetime, int, Tuple[str], Callable, Enum, List[str]]
    initial: str = "anything"
    number_of_repeat: int = 1
    disp_value: str = ""
    required: bool = True
    obscured: bool = False
    def __str__(self) -> str:
        return f"{self.disp_value}"
    def __repr__(self) -> str:
        return f"{self.disp_value}"
    def __post_init__(self):
        self.create_disp_value()
    def create_disp_value(self):
        if not self.value and str(self.value) != "0":
            self.disp_value = f"{self.initial}=Noneapi.params.103"
            return
        # if isinstance(self.value, Enum) :
        if hasattr(self.value, "value"):
            self.disp_value = f"{self.initial}={self.value.value}"
        if isinstance(self.value, (tuple, list,)):
            value: str = "-".join(self.value)
            self.disp_value = f"{self.initial}={value}"
            return
        if isinstance(self.value, (str, int,)):
            self.disp_value = f"{self.initial}={self.value}"
            return
        if not self.value:
            if self.required:
                raise f"{self.__class__.__name__} parameter is required"
            return
        self.disp_value = f"{self.initial}={self.value}"
    def reportable_value(self):
        if self.obscured:
            return f"{self.initial}=obscured_for_report"
        return self.disp_value
    def explanation(self):
        if self.obscured:
            return f"{self.initial}: obscured_for_report"
        if callable(self.value):
            disp_value = f"{self.initial}: {self.value.__name__}"
            return disp_value
        if isinstance(self.value, (list, str,)):
            disp_value = f"{self.initial}: {', '.join(self.value)}"
            return disp_value
        return f"{self.initial}: {self.value}"
from typing import Callable
@dataclass
class Apikey(UrlParam):
    initial: str = "key"
    obscured: bool = True
class DecimalSeparator(UrlParam):
    value: Separator = Separator.period
    initial: str = "DecimalSeparator"
@dataclass
class Series(UrlParam):
    """ Frequency """
    value: Union[List[str], Tuple[str]] = ()
    initial: str = "series"
@dataclass
class DataGroup(UrlParam):
    """ Frequency """
    value: str = ''
    initial: str = "datagroup"
    required: bool = True
    def create_disp_value(self):
        self.disp_value = f"{self.initial}={self.value}"
@dataclass
class Frequency(UrlParam):
    """ Frequency """
    # value: Union[str, datetime, int, Tuple, Callable, Enum]
    initial: str = "frequency"
    required: bool = False
@dataclass
class Formulas(UrlParam):
    """ Formulas """
    initial: str = "formulas"
    required: bool = False
    def create_disp_value(self):
        repeat = "-".join([str(self.value) for x in range(self.number_of_repeat)])
        self.disp_value = f"{self.initial}={repeat}"
@dataclass
class Aggregations(UrlParam):
    """ Aggregations """
    initial: str = "aggregationTypes"
    required: bool = False
    def create_disp_value(self):
        repeat = "-".join([str(self.value) for x in range(self.number_of_repeat)])
        self.disp_value = f"{self.initial}={repeat}"
@dataclass
class DateStart(UrlParam):
    """ DateStart """
    value: str = ""
    initial: str = "startDate"
@dataclass
class DateEnd(UrlParam):
    """ DateEnd """
    value: str = ""
    initial: str = "endDate"
@dataclass
class dataTypeParam(UrlParam):
    value: dataTypeEnum = dataTypeEnum.csv
    required = True
    initial: str = "type"
    def create_disp_value(self):
        self.disp_value = f"{self.initial}={self.value.value}"
@dataclass
class ModeParamDataGroups(UrlParam):
    initial: str = "mode"
    required = True
    # def create_disp_value(self):
    #     self.disp_value = f"{self.initial}={self.value.value}"
@dataclass
class CodeParamDataGroups(UrlParam):
    value = 0
    initial: str = "code"
    required = True
    def create_disp_value(self):
        if hasattr(self.value, "value"):
            value = self.value.value
        else:
            value = self.value
        self.disp_value = f"{self.initial}={value}"
def convert_list_params_to_str(alist: Union[List[UrlParam], Tuple[UrlParam]], fnc: Union[Callable, str] = str):
    separator_char = "&"
    if fnc == str:
        alist = [str(x) for x in alist if x is not None]
        alist_str: str = separator_char.join(alist)
    else:
        indent = " " * 10
        block = "-" * 25
        separator_char = "\n" + indent
        if fnc == "report":
            alist = [x.reportable_value() for x in alist if x.reportable_value() is not None]
        if fnc == "explanation":
            alist = [x.explanation() for x in alist if x.explanation() is not None]
        alist_str: str = block + f"{fnc}" + block + "\n" + indent + separator_char.join(alist)
    return alist_str



from .index_classes import GeneralIndexesDatagroupIndividual, \
    GeneralIndexesDatagroupSeriesList
from .index_util_funcs import json_to_excel, json_to_df, make_df_float
from ..components.api_params import DateStart, DateEnd
from ..components.options_class import SingletonOptions
from ..config.config import ConfigBase
from ..initial.start_options import default_data_folder_name, Default_Prefix_
from ..requests_.ev_request import EVRequest
config = ConfigBase()
# EVDSApiDomainDatagroupIndividual
def try_to_make_excel(json_content, file_name, try_float):
    try:
        json_to_excel(json_content, file_name, try_float=try_float)
        from ..common.colors import print_with_success_style
        print_with_success_style(f"{file_name} was created...")
    except:
        from ..common.colors import print_with_failure_style
        print_with_failure_style(f"{file_name} could not be created...")
def get_datagroup_individual_with_code(code_str: str):
    """get_datagroup_individual_with_code"""
    json_content = get_datagroup_individual_with_code_helper(code_str)
    # file_name = rf"SeriesData\EVPY_Data_{code_str}.xlsx"
    file_name = rf"{default_data_folder_name}\{Default_Prefix_}{code_str}.xlsx"
    if not json_content:
        return False
    if 'items' in json_content:
        json_content = json_content['items']
    try_to_make_excel(json_content, file_name, try_float=True)
    return json_content
def get_datagroup_individual_with_code_helper(
        code_str: str,
        start_date: str = None,
        end_date: str = None
):
    """get_datagroup_individual_with_code_helper"""
    gid = GeneralIndexesDatagroupIndividual(code=0, EVRequest_=EVRequest(
        options_=SingletonOptions()))  # this number will be overwritten . does not matter what number
    gid.create_url_first_part()
    gid.add_extra_params(code=code_str)
    if not start_date:
        start_date = SingletonOptions().get_valid_value("default_start_date")
    if not end_date:
        end_date = SingletonOptions().get_valid_value("default_end_date")
    date_start: DateStart = DateStart(value=start_date)
    date_end: DateEnd = DateEnd(value=end_date)
    gid.complete_url_instance.add_item(date_start)
    gid.complete_url_instance.add_item(date_end)
    gid.complete_url_instance.refresh_url()
    gid.complete_url_instance.add_apikey()  # this will get apikey itself if None given
    json_content = gid.get_json()
    return json_content
def get_series_list_of_subject(code_str: str):
    # https://evds2.tcmb.gov.tr/service/evds/serieList/key=XXXXX&type=csv&code=bie_yssk
    gid = GeneralIndexesDatagroupSeriesList(code=0, EVRequest_=EVRequest(
        options_=SingletonOptions()))  # this number will be overriten . does not matter what number
    gid.create_url_first_part()
    gid.add_extra_params(code=code_str)
    gid.complete_url_instance.refresh_url()
    json_content = gid.get_json()
    # file_name = rf"SeriesData\EVPY_Data_{code_str}_EXPLANATION.xlsx"
    file_name = rf"{default_data_folder_name}\{Default_Prefix_}{code_str}_EXPLANATION.xlsx"
    try_to_make_excel(json_content, file_name, try_float=False)
    return json_content
# ----------------------------------------------------------------------------------
""" Functions to return DF """
"""
    start_date = SingletonOptions().get_valid_value("default_start_date")
    end_date = SingletonOptions().get_valid_value("default_end_date")
"""
def get_df_datagroup(
        datagroup: str,
        start_date: str = None,
        end_date: str = None
):
    """returns all series as df to extend
        params:
        -----------------------
        datagroup : str
            e.g. `bie_yssk`
        start_date : str
            e.g. `31-01-2010`
        end_date : str
            e.g. `31-01-2030`
    """
    # https://evds2.tcmb.gov.tr/service/evds/datagroup=bie_yssk&startDate=01-06-2017&endDate=07-09-2017&type=csv&key=XXXX
    json_content = get_datagroup_individual_with_code_helper(datagroup, start_date, end_date)
    df = json_to_df(json_content)
    df = make_df_float(df)
    return df



from .help_messages import *
from .cmd_formats import *
api_key_not_set_msg = """
        ---------------------------------------------------------------
               WARNING
        ---------------------------------------------------------------
        Api key not set .
        ---------------------------------------------------------------
        you may continue with saving your api key by answering
                 `yes`
        Program will store your api key in your work environment creating a folder `APIKEY_FOLDER` and read it when
        it needs to request data from EVDS server.
        This key will be used when you run your requests and will be saved to your work environment
        in a folder named APIKEY_FOLDER in a text file not as a plane text but in a obscured way.
        If you are working in a public workspace we suggest that you delete this folder once you are finished making requests.
        Program will not use or store your API key to any other folder therefore when you delete, remove or rename this folder
        program will ask your API key again.
        Below, there are links of the institution's web service. Signing up or logining in you may find your API key on
        your account.
        You may get your api key from your account on CBRT EVDS website.
        CBRT EVDS API (TCMB EVDS API)
        -----------------
        Sign up / Log in
        https://evds2.tcmb.gov.tr/index.php?/evds/login
        Main page
        https://evds2.tcmb.gov.tr
        Docs
        https://evds2.tcmb.gov.tr/index.php?/evds/userDocs
        If you already have an API key may continue with saving your api key by answering
                 `yes` now and then your API key"""
you_need_to_run_setup_options = f"""
    it looks like options.py not created.
    try command below from command line in your project's root folder.
{setup_now_command}
"""
copying_options_msg = "Found an options file... Trying to copy it to temporary env."
series_file_not_found_error_msg = """
create_series_file
"""



from pathlib import Path
import os
import base64
api_key_folder_name = Path() / "APIKEY_FOLDER"
api_key_file_name = api_key_folder_name / "api_key.txt"
def decode(base64_bytes) -> str:
    if isinstance(base64_bytes, str):
        base64_bytes = str.encode(base64_bytes)
    str_bytes = base64.b64decode(base64_bytes)
    decoded = str_bytes.decode("ascii")
    return decoded
def encode(text: str) -> bytes:
    t_bytes = text.encode("ascii")
    encoded = base64.b64encode(t_bytes)
    return encoded
def WriteBytes(file_name: str, content_bytes: bytes):
    with open(file_name, 'wb') as f:
        f.write(content_bytes)
def ReadBytes(file_name: str):
    with open(file_name, "rb") as f:
        return f.read()
from evdspy.EVDSlocal.config.apikey_class import ApikeyClass
# def check_key_before_saving(instance, api_key_answer: str):
#     ApikeyClass().now_testing_is_key_is_valid = api_key_answer
#     return instance.check_api_key_with_simple_Wrapper()
def save_api_key_to_file(instance, api_key_answer):
    instance.wait(1)
    if not api_key_folder_name.is_dir():
        os.makedirs(api_key_folder_name)
    # api_key_answer = instance.get_input("Your api key : ")
    # Write(api_key_file_name, str(encode(api_key_answer)), "")
    WriteBytes(str(api_key_file_name), encode(api_key_answer))
    # test for file api key for runtime (temp)
    ApikeyClass().set_api_key_filetype(value=api_key_answer)
    # ApikeyClass().now_testing_is_key_is_valid = api_key_answer
    print("your api file was saved...")
    instance.wait(1)
    # instance.set_apikey(api_key_answer)
    api_key_from_file = get_api_key_from_file_improved()
    return api_key_from_file
def get_api_key_from_file_improved():
    if not api_key_file_name.is_file():
        return False
    def read_api_file():
        content = ReadBytes(str(api_key_file_name))
        return content
    pot_key = read_api_file()
    if pot_key:
        return decode(pot_key)
    return False
def check_api_key_on_load():
    """ not compatible 3.8-"""
    # if api_key := get_api_key_from_file_improved():
    #     ApikeyClass().set_api_key_filetype(value=api_key)
    """changing to this"""
    api_key = get_api_key_from_file_improved()
    if api_key:
        ApikeyClass().set_api_key_filetype(value=api_key)



from ..common.common_imports import *
from ..common.files import *
from dataclasses import dataclass , field 
from evdspy.EVDSlocal.config.config import *
from ..config.apikey_class import ApikeyClass
from ..components.options_class import Options, load_options, SingletonOptions, read_user_options_on_load
from evdspy.EVDSlocal.initial.start_options import Default_Prefix_
from evdspy.EVDSlocal.components.bucket_from_series import BucketFromSeriesFile, null_BucketFromSeriesFile
from evdspy.EVDSlocal.components.url_class import URLClass
from evdspy.EVDSlocal.components.api_params import Series, DateStart, DateEnd, dataTypeParam, dataTypeEnum
from typing import Tuple , Optional 

# ------------------------------------------------------------------------------
# -------------------------------------------------EvdsSeri----------------------
@dataclass
class EvdsSeri:
    """
    EvdsSeri
    """
    ID: str
    bfs: BucketFromSeriesFile
    name: str = ""
    subject: str = ""
    def __repr__(self):
        content = f"""
        <EvdsSeri>
            ID : {self.ID}
            name : {self.name}
            subject : {self.subject}
            bfs : {self.bfs}
        </EvdsSeri>
"""
        # print(content)
        return content
#   ----------------------------------------------------------    / EvdsSeri
liste = ('TP.IHRACATBEC.9999', 'TP.IHRACATBEC.31', 'TP.IHRACATBEC.41')
seri_evds_test_objs = tuple(EvdsSeri(x, bfs=null_BucketFromSeriesFile) for x in liste)
# ------------------------------------------------------------------------------
# /*
#               EvdsSeriesRequest
#
# ------------------------------------------------------------------------------
# */
@dataclass
class EvdsSeriesRequest:
    bfs: BucketFromSeriesFile
    options_: Optional[Options] = load_options()
    # series_list: field(default_factory=list) = field(default_factory=list[seri_evds_test_objs])
    series_list: field(default_factory=list) = tuple(seri_evds_test_objs)
    complete_url_instance: URLClass = None
    def __post_init__(self):
        read_user_options_on_load()
        self.start_date: str = SingletonOptions().get_valid_value("default_start_date")
        self.end_date: str = SingletonOptions().get_valid_value("default_end_date")
        self.update_url_instance()
    def update_url_instance(self):
        """_summary_
        update_url_instance
        """
        date_start: DateStart = DateStart(value=self.start_date)
        date_end: DateEnd = DateEnd(value=self.end_date)
        type_param = dataTypeParam(value=dataTypeEnum.csv)
        slist: Tuple[str] = tuple(x.ID for x in self.series_list)
        series_instance = Series(slist)
        full_list = [series_instance, date_start, date_end, type_param]
        complete_url_instance = URLClass(full_list)
        number_of_series = len(slist)
        # ADD other params from file
        complete_url_instance.add_formulas(self.bfs.formulas, number_of_series)
        complete_url_instance.add_frequency(self.bfs.frequency)
        complete_url_instance.add_aggtype(self.bfs.aggregateType, number_of_series)
        api_key = ApikeyClass().get_valid_api_key()
        complete_url_instance.add_apikey(api_key)
        self.complete_url_instance = URLClass(complete_url_instance.url_items)
        self.complete_url_instance.refresh_url()
        self.complete_url_instance.create_report()
#   ----------------------------------------------------------    / EvdsSeriesRequest
test_series_ = EvdsSeriesRequest(options_=load_options(), series_list=seri_evds_test_objs,
                                 bfs=null_BucketFromSeriesFile)
# ------------------------------------------------------------------------------
# /*
#               EvdsSeriesRequestWrapper(s)
#
# ------------------------------------------------------------------------------
# */
@dataclass
class EvdsSeriesRequestWrapper():
    name: str
    subject: str
    EvdsSeriesRequest_: EvdsSeriesRequest
    bfs: BucketFromSeriesFile = None  # field(default=null_BucketFromSeriesFile)
    # bfs: BucketFromSeriesFile = null_BucketFromSeriesFile
    def __post_init__(self):
        self.EvdsSeriesRequest_.bfs = self.bfs
def EvdsSeriesRequestWrapperBasic(name: str, subject: str, EvdsSeriesRequest_: EvdsSeriesRequest):
    bfs = BucketFromSeriesFile(name, subject, prefix=Default_Prefix_)
    return EvdsSeriesRequestWrapper(name, subject, EvdsSeriesRequest_, bfs)
def EvdsSeriesRequestWrapperFromFile(bfs: BucketFromSeriesFile, EvdsSeriesRequest_: EvdsSeriesRequest):
    return EvdsSeriesRequestWrapper(bfs.subject, bfs.subject, EvdsSeriesRequest_, bfs)
#   ----------------------------------------------------------    / EvdsSeriesRequestWrapper
__all__ = [
    'EvdsSeriesRequest',
    'test_series_',
    'EvdsSeri',
    'EvdsSeriesRequestWrapper',
    'EvdsSeriesRequestWrapperFromFile',
    'EvdsSeriesRequestWrapperBasic',
]



from ..index_requests.categories import get_categories
# from ..index_requests.datagroups import get_datagroups

def test_get_categories(capsys):
    get_categories()
# def test_get_datagroups(capsys):
#     get_datagroups()




from abc import ABC   
###########################  Base Exception Class ###########################
class BaseException(Exception):
    def __init__(self, message="Something happpened"):
        self.message = message
        super().__init__(self.message)
    def __str__(self):
        return f' -> {self.message}'
###########################  / Base Exception Class ###########################
class ApiKeyNotSetError(BaseException):
    def __init__(self, message="Api key not set yet."):
        self.message = message
class SeriesEmptyError(BaseException):
    def __init__(self, message="Series cannot be empty"):
        self.message = message
class OptionsFileNotLoadedError(BaseException):
    def __init__(self, message="Options file not loaded..."):
        self.message = message
class SeriesFileDoesNotExists(BaseException):
    def __init__(self, message="Series file does not exist..."):
        self.message = message
class BucketFromSeriesFolderCreateError(BaseException):
    def __init__(self, message="BucketFromSeriesFolderCreateError folder problem"):
        self.message = message
class ExceptionMixing(ABC):
    ...


from .dotenv import load_env
from ..common.common_imports import *
from dataclasses import dataclass
from typing import Union
from enum import Enum, auto
import os
import traceback

# try :
#
#     from dotenv import load_dotenv
#     load_dotenv()
# except Exception:
#
#     print("""[solution]
#           $ pip install python-dotenv
#           """)
#     traceback.print_exc()
#     raise ModuleNotFoundError()
from ..config.config import config

from ..messages.error_classes import ApiKeyNotSetError

"""
    ========================== load_env ==========================
"""
load_env()





class ApiKeyType(Enum):
    runtime = auto()
    from_file = auto()
    from_file_options = auto()


@dataclass
class ApiKeyValue:
    name: str
    value: Union[str, bool] = False
    type_: ApiKeyType = ApiKeyType.runtime


def dot_get_env_evds():
    api_key_env = os.getenv("EVDS_API_KEY")
    api_key_env = api_key_env if len(str(api_key_env)) > 5 else False
    return ApiKeyValue("dotenv", api_key_env)


# assert dot_get_env_evds().value


class ApiKeyDict:
    runtime_apikey: ApiKeyValue
    from_file_apikey: ApiKeyValue
    from_file_options_api_key: ApiKeyValue
    dotenv_api_key: ApiKeyValue

    def __init__(self):
        self.dotenv_api_key: ApiKeyValue = dot_get_env_evds()
        self.runtime_apikey = ApiKeyValue("runtime", False, ApiKeyType.runtime)
        self.from_file_apikey = ApiKeyValue("fromfile", False, ApiKeyType.from_file)
        self.from_file_options_api_key = ApiKeyValue(
            "fromfile_options", False, ApiKeyType.from_file_options
        )

    def set_value(self, type_: Union[str, ApiKeyType], value):
        if type_ in ["runtime", ApiKeyType.runtime]:
            self.runtime_apikey.value = value
            return
        if type_ in ["fromfile", ApiKeyType.from_file]:
            self.from_file_apikey.value = value
            return
        self.from_file_options_api_key = value


import time


class ApikeyClass(object):
    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(ApikeyClass, cls).__new__(cls)
            cls.post_init(cls)
        return cls.instance

    def __repr__(cls):
        return f"""
   cls.now_testing_is_key_is_valid: {cls.now_testing_is_key_is_valid:}
   cls.current_key_using : {cls.current_key_using}
   cls.APIKEYDict : {cls.APIKEYDict}
"""

    def post_init(cls):
        cls.APIKEYDict: ApiKeyDict = ApiKeyDict()
        cls.now_testing_is_key_is_valid: Union[str, bool] = False
        cls.current_key_using: Union[str, bool] = cls.APIKEYDict.runtime_apikey.value

    def no_api_msg(self):
        template = """
        Api Key not set yet.
        --------------------------------------------
        how to set api key?
        --------------------------------------------
        Option 1 (from console)
            $ evdspy save
        Option 2 (python console with menu)
            python> from evdspy import menu
            python> menu()
            Selection 10
            10 | SAVE API KEY TO FILE
        Option 3 (python console )
            python> from evdspy import save
            python> save()
        """
        print(template)

    def get_valid_api_key(cls, check=False):
        if cls.now_testing_is_key_is_valid:
            return cls.now_testing_is_key_is_valid
        key_objs = cls.get_api_keys()
        if not key_objs:
            print(key_objs)
            
            if check:
                cls.no_api_msg()
                time.sleep(4)
                raise ApiKeyNotSetError("No valid API keys found")
            if config.current_mode_is_test:
                return "VALID_KEY_FOR_TEST"
            return False
        return key_objs[0].value

    @property
    def key(self):
        if config.current_mode_is_test:
            return "VALID_KEY_FOR_TEST"
        key = self.get_valid_api_key(check=True)
        return key

    @staticmethod
    def obscure(key=None):
        from evdspy.EVDSlocal.utils.utils_general import encode

        if key is None:
            key = ApikeyClass().key
        if not isinstance(key, str):
            return "..."

        strings = []
        for index, i in enumerate(str(encode(key))):
            if index < 7 or index % 3 == 0:
                strings.append("*")
            else:
                strings.append(i)
        return "".join(strings)

    def set_api_key_runtime(cls, value: str):
        cls.instance.APIKEYDict.set_value(ApiKeyType.runtime, value)
        return cls.instance.APIKEYDict.runtime_apikey.value

    def set_api_key_filetype(cls, value: str):
        cls.instance.APIKEYDict.set_value(ApiKeyType.from_file, value)
        return cls.instance.APIKEYDict.from_file_apikey.value

    def set_api(cls, key: str):
        print("set api not implemented")
        raise "set api not implemented"

    def get_api_keys(cls):

        keys = [
            # dot_get_env_evds(),
            cls.instance.APIKEYDict.dotenv_api_key,
            cls.instance.APIKEYDict.runtime_apikey,
            cls.instance.APIKEYDict.from_file_apikey,
            cls.instance.APIKEYDict.from_file_options_api_key,
        ]
        return [x for x in keys if x.value is not False]

    def get_api_keys_dict(cls):
        return cls.instance.APIKEYDict

    def get_api_key_runtime(cls):
        return cls.instance.APIKEYDict.runtime_apikey.value

    def get_api_key_fromfile(cls):
        return cls.instance.APIKEYDict.from_file_apikey.value

    def get_api_key_fromfile_options(cls):
        return cls.instance.APIKEYDict.from_file_options_api_key



from evdspy.EVDSlocal.components.evds_files import EvdsSorguSeries
from evdspy.EVDSlocal.components.evds_seri_files import EvdsSeri, EvdsSeriesRequest, EvdsSeriesRequestWrapper
from evdspy.EVDSlocal.series_format.populate_series import PopulateSeries
from evdspy.EVDSlocal.series_format.series_creator import SeriesFileFormat, create_series_file_from_Wrapper, \
    create_series_text_file
from evdspy.EVDSlocal.initial.start_args import Args
from evdspy.EVDSlocal.components.options_class import *
from evdspy.EVDSlocal.components.evds_seri_files import test_series_
from ..common.table import Table_
from ..initial_setup.setup_folders import check_setup
from ..messages.error_classes import *
from evdspy.EVDSlocal.log_classes.log_template import deb
from evdspy.EVDSlocal.config.config import ConfigBase
from evdspy.EVDSlocal.manual_requests.prepare import basic_for_test
from evdspy.EVDSlocal.config.apikey_class import ApikeyClass
from evdspy.EVDSlocal.utils.utils_general import *
from ..common.colors import print_with_info_style
import typing as t
from dataclasses import dataclass  , field 
from typing import Optional ,List 
from .cache_folder import get_cache_folder
@dataclass
class LoadModulesClass():
    options_: Optional[any] = field(default_factory=load_options)
    args: Optional[Args] = field(default_factory=Args)  # Args(sys.argv)
    evds_list: field(default_factory=list) = field(default_factory=list)
    series_filename: str = default_series_file_name
    api_key_test: bool = False
    item_pointer = 0
    # TODO
    """
    """
    def placebo_item(self):
        evds = EvdsSorguSeries(
            options_=self.options_,
            session=None,
            subject="placebo",
            series_=test_series_,
            file_name="placebo",
            args=self.args)
        return evds
    def summary(self):
        if not self.evds_list or not len(self.evds_list) > 0:
            print("No EvdsSorguSeries yet")
            return
        for item in self.evds_list:
            item.summary()
    def series_from_file_test(self):
        """Reads series' code from given file name inside initial folder default :`series.txt` """
        return self.create_EVDSSorguSeries()
    def main(self):
        return self.series_from_file()
    def series_from_file(self):
        """ MAIN function """
        result = []
        try:
            result: t.List[EvdsSorguSeries] = self.series_from_file_core()
            # _ = list(x.display for x in result)
        except SeriesFileDoesNotExists:
            print(SeriesFileDoesNotExists().message)
        return result
    def convert_wrapperlist_to_text(self, wrapper_list: List[EvdsSeriesRequestWrapper]):
        content = create_series_file_from_Wrapper(wrapper_list)
        return content
    def create_locked_series_file(self):
        wrapper_list = self.populate_evds_wrapperList()
        content = self.convert_wrapperlist_to_text(wrapper_list)
        file_name = default_series_file_name.split(".cfg")[0] + "-locked.cfg"
        create_series_text_file(file_name, content)
    def create_EVDSSorguSeries(self):
        return self.populate_evds_list()
    def populate_evds_wrapperList(self):
        # return
        # if self.api_key_test:
        #     """ api key already was tested here we are checking if it is valid """
        #     self.series_filename = str(
        #         ConfigBase().runtime_file_name_root_path / "series_format" / "series_test_runtime.cfg")
        #     # self.series_filename = "series_format/series_test_runtime.cfg"
        sff = SeriesFileFormat(self.series_filename)
        return sff.EvdsSeriesRequestWrapperList
    def check_for_apikey_with_simple_wrapper(self, api_key):
        basic_for_test(api_key)
        exit()
        self.api_key_test = True
        self.evds_list = self.populate_evds_list()
        if not self.evds_list:
            return False
        self.evds_list = [self.evds_list[0]]
        return self.evds_to_excel(api_key_test=True)
        # return len(evds_list) and isinstance(evds_list[0], EvdsSorguSeries)
    def populate_evds_list(self):
        """==>>>>   (Deep )   PopulateSeries() =>
                    (Surface) SeriesFileFormat .EvdsSeriesRequestWrapper_
        Reads series' code from given file name
        inside initial folder default :`series.txt`
        returns a list of wrapper
        ( EvdsSeriesRequestWrapper1 ,  EvdsSeriesRequestWrapper2 , ... )
        EvdsSeriesRequestWrapper :
            name
            subject
            EvdsSeriesRequest
        """
        list_ = []
        EvdsSeriesRequestWrapperList = self.populate_evds_wrapperList()
        for index, EvdsSeriesRequestWrapper_ in enumerate(EvdsSeriesRequestWrapperList):
            evds = EvdsSorguSeries(
                options_=self.options_,
                session=None,
                subject=EvdsSeriesRequestWrapper_.subject,
                series_=EvdsSeriesRequestWrapper_.EvdsSeriesRequest_,
                file_name=EvdsSeriesRequestWrapper_.name,
                args=self.args,
                bfs=EvdsSeriesRequestWrapper_.bfs,
            )
            list_.append(evds)
        return list_
    def evds_to_excel(self, api_key_test=False):
        for evds in self.evds_list:
            try:
                evds.save_excel()
                evds.report()
            except Exception as exc:
                # print(exc)
                # print(evds)
                ...
        return True
    # def display_items(self):
    #     _ = (x.display for x in self.evds_list)
    #     print(_)
    def series_from_file_core(self):
        self.evds_list = self.create_EVDSSorguSeries()
        self.display_items(self.evds_list)
        # print(self.evds_list )
        self.evds_to_excel()
        self.create_locked_series_file()
        return self.evds_list
    def display_messages(self, *msgs, wait_num=None):
        if wait_num is None:
            wait_num = {0: 3, 1: 2, 2: 1}
        def print_and_wait():
            for index, msg in enumerate(msgs):
                print_with_info_style(msg)
                time.sleep(wait_num.get(index, 2))
        print_and_wait()
    def display_reading_message(self, len_):
        msg2 = """
        collecting...
"""
        msg3 = f"""
        collected {len_} items.
"""
        wait_num = {0: 4, 1: 3, 2: 3}
        self.display_messages(msg2, msg3, wait_num=wait_num)
    def display_items(self, evds_list: t.List[EvdsSorguSeries]):
        msg1 = """
            Now reading `config_series.cfg` file...
        """
        self.display_messages(msg1)
        if not evds_list:
            if not self.evds_list:
                evds_list = ()
            else:
                evds_list = self.evds_list
        if isinstance(evds_list, bool):
            print(evds_list)
            self.display_reading_message(0)
            return
        if not evds_list:
            return
        self.display_reading_message(len(evds_list))
        for index, item in enumerate(evds_list):
            item.display(index + 1)
            time.sleep(1)
    def seri_ornek(self, options_, args):
        liste = [
            'TP.IHRACATBEC.9999',
            'TP.IHRACATBEC.31',
            'TP.IHRACATBEC.41',
        ]
        seri_evds_objs: list[EvdsSeri] = [EvdsSeri(x) for x in liste]
        series_ = EvdsSeriesRequest(self.options_, seri_evds_objs)
        file_name = "test34"
        evds = EvdsSorguSeries(
            session=None,
            series_=series_,
            file_name=file_name
        )
        evds.save_excel(evds.file_name)
    def check(self):
        evds_items = self.series_from_file_test()
        if len(evds_items):
            self.check_items(evds_items)
    def check_items(self, evds_items):
        config = ConfigBase()
        options_copy_file_name = config.runtime_file_name_path
        user_options_file_name = config.user_options_file_name
        ps = PopulateSeries()
        series_file_was_created = False
        f = Path(ps.input_file_name)
        if f.is_file():
            series_file_was_created = True
        folders_ok = False
        if check_setup():
            folders_ok = True
        api_key_is_ok = False
        if ApikeyClass().get_valid_api_key():
            api_key_is_ok = True
        def display():
            options_file_name = "options.cfg"
            options_file_created = Path(options_file_name).is_file()

            cache_folder_ok = get_cache_folder().is_dir()


            options_display = SingletonOptions().check()
            if 'NEXT_RELEASE' not in str(Path.cwd()):
                workspace = str(Path.cwd()).replace(":", ".")
            else:
                workspace = r'test_area3'
            hiddenok = "hidden"
            if not api_key_is_ok:
                hiddenok = ""
            # [{str(get_cache_folder()).replace(":", " ")}]
            msg = f"""
Workspace : {workspace}
Folders created          :{folders_ok}
Cache folder was created :{cache_folder_ok}  
Series file was created  :{series_file_was_created} {indent} {ps.input_file_name}
Options file was created :{options_file_created} {indent} {options_file_name}
Api key was set          :{api_key_is_ok} {indent} {hiddenok}
        {options_display}
                """
            # print(msg)
            Table_().show(content=msg,
                          title=" Checking installation and other setup requirements ",
                          columns=('', '',),
                          skiprow=1)
        display()




# ------------------------------------------------------------------------------
import pandas as pd
from evdspy.EVDSlocal.initial.start_options import default_series_file_name
from evdspy.EVDSlocal.config.config import *
from evdspy.EVDSlocal.components.excel_class import *
from ..requests_.request_error_classes import Internal
from ..utils.utils_general import *
from ..requests_.my_cache import MyCache
from ..components.options_class import Options, load_options
from ..components.evds_seri_files import EvdsSeriesRequest
from evdspy.EVDSlocal.requests_.ev_request import EVRequest
from ..config.credentials_file import Credentials
from ..initial.start_args import *
from ..config.apikey_class import *
from dataclasses import dataclass
from typing import Tuple
from ..common.colors import *
from abc import ABC 

import os 
# ------------------------------------------------------------------------------
m_cache = MyCache()
number = 0
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# /*
#               EvdsSorgu
#
# ------------------------------------------------------------------------------
# */
from evdspy.EVDSlocal.components.evds_seri_files import BucketFromSeriesFile
from evdspy.EVDSlocal.components.request_or_cache import RequestOrCacheResultInfo
from evdspy.EVDSlocal.state.current_state import CurrentState
from evdspy.EVDSlocal.components.url_class import URLClass
current_state = CurrentState()
class DfColumnsDoesNotMatch(BaseException):
    """DfColumnsDoNotMatch"""
from evdspy.EVDSlocal.messages.error_classes import ApiKeyNotSetError
# field(default_factory=URLClass)
@dataclass
class EvdsSorgu(ABC):
    """ EVDS Genel SÃ„Â±nÃ„Â±f"""
    options_: Options
    session: any
    series_: EvdsSeriesRequest = field(
        default_factory=EvdsSeriesRequest)  # field(default=test_series_)  # None  # test_series_
    file_name: Optional[str] = "test"
    subject: str = "test_subject"
    df: any = None
    excel_out_ext: str = "xlsx"
    args: Args = field(default_factory=Args)  # Args(sys.argv)
    url: Optional[str] = None
    proxy: Optional[str] = False
    bfs: BucketFromSeriesFile = None
    df_checked: bool = False
    request_or_cache_resultInfo_: RequestOrCacheResultInfo = field(default_factory=RequestOrCacheResultInfo)
    def __post_init__(self):
        # self.URL_Instance = URLClass()
        self.credentials = field(default_factory=Credentials)  # Credentials(options_=self.options_)
        self.address: str = ''  # field(default=self.credentials.address)
        self.EVRequest_: EVRequest = field(default_factory=EVRequest)  # EVRequest(options_=self.options_, session=None)
        self.parent_folder = Path(__file__).parent
        self.core_file_name = self.file_name
        self.file_name = str(Path().absolute() / "output" / self.file_name)
        self.current_file_created = ""
        self.display()
        self.assign_results_to_state()
        if not self.check_api_key_first():
            print_with_failure_style("Api key")
    def assign_results_to_state(self):
        """
        save result stats to state
            to finally access from bfs
        """
        current_state.set_result_dict(str(id(self.bfs)), self.request_or_cache_resultInfo_)
    def report(self):
        m_cache.report()
        ...
    def display(self, *args):
        # global number
        # number += 1
        items_series = (x.ID for x in self.series_.series_list)
        dec = f"""
----------------------------------------
    Order item from '{default_series_file_name}'
    Details :
    bfs : {self.bfs}
___________________
    filename: {self.core_file_name}
    subject : {self.subject}
    series :
        {", ".join(items_series)}
----------------------------------------
"""
        print(dec)
        return dec
    def check_api_key_first(self):
        # TODO  ApikeyClass().get_valid_api_key()
        if not self.EVRequest_:
            self.EVRequest_ = EVRequest(options_=load_options(),
                                        session=None,
                                        proxy=None,
                                        args=Args(())
                                        )
        try:
            if ApikeyClass().get_valid_api_key():
                # self.EVRequest_.check_any_apikey()
                return True
        except ApiKeyNotSetError:
            return False
    def create_url(self):
        api_key = ApikeyClass().get_valid_api_key()
        if not config.current_mode_is_test:
            assert api_key is not False, "api_key is False "
        self.complete_url_instance: URLClass = self.series_.complete_url_instance
        self.request_or_cache_resultInfo_.url_instance = self.series_.complete_url_instance
        self.complete_url_instance.refresh_url()
        self.request_or_cache_resultInfo_.safe_url = f"{self.complete_url_instance.create_report()}"
        will_be_requested: Tuple = (
            self.complete_url_instance.url,
            self.complete_url_instance.url_only_required,
        )
        return will_be_requested
    def get_json(self, url: str):
        """ Not implemented """
        raise NotImplementedError
        # return self.EVRequest_.get_request_before_cache(f"{url}&type=json").json()
    def get_csv(self, url: str):
        if not hasattr(self.EVRequest_, 'URL_Instance'):
            self.EVRequest_ = EVRequest(options_=load_options())
            self.EVRequest_.URL_Instance = URLClass()
        self.EVRequest_.URL_Instance = self.complete_url_instance
        result = self.EVRequest_.get_request_before_cache(f"{url}&type=csv")
        if hasattr(result, "text"):
            self.csv = result.text
            return result.text
        return False
    @staticmethod
    def extract_values_from_line(line: str, sep=","):
        return line.split(sep)
    @staticmethod
    def line_gen(buffer: str):
        if not buffer:
            return False
        for item in buffer.split("\n"):
            yield item
    # @staticmethod
    def get_columns_from_first_row(self):
        first_row = self.csv.split("\n")[0].split(",")
        return first_row
    def get_columns(self, df):
        current_columns = df.columns  # rangeIndex
        if isinstance(current_columns, pd.RangeIndex):
            current_columns_len = len([x for x in current_columns])
        else:
            current_columns_len = len(current_columns)
        potential_columns = self.get_columns_from_first_row()
        if len(potential_columns) != current_columns_len:
            raise DfColumnsDoesNotMatch()
        df.columns = self.get_columns_from_first_row()
        self.request_or_cache_resultInfo_.columns = df.columns
        UnixTime = 'UNIXTIME'
        if UnixTime in df.columns:
            df.drop(UnixTime, axis=1, inplace=True)
        if config.current_mode_is_test:
            self.save_excel("testpytest2.xlsx")
        return df
    @staticmethod
    def list_to_df(items):
        return pd.DataFrame(items)
    def make_df_float(self):
        def check_cols(x: str):
            items = (
                "Tarih" not in x,
                "WEEK" not in x,
                "DAY" not in x,
            )
            return all(items)
        columns = (x for x in self.df.columns if check_cols(x))
        other_cols = (x for x in self.df.columns if not check_cols(x))
        old_df = self.df
        try:
            self.df = self.df[columns].astype(float)
            for other_column in other_cols:
                self.df[other_column] = old_df[other_column]
        except:
            # print_with_failure_style("Could not convert some columns to float type...")
            ...
    def convert_csv_df(self, csv_buffer: str):
        if not csv_buffer:
            return False
        pd_rows = list(
            map(self.extract_values_from_line,
                (item for item in self.line_gen(csv_buffer))
                )
        )
        pd_rows.pop(0)
        self.df_no_columns = self.list_to_df(pd_rows)
        try:
            self.df = self.get_columns(self.df_no_columns)
        except DfColumnsDoesNotMatch:
            print_with_failure_style("columns did not match returning df with no columns...")
            self.df = self.df_no_columns
            print(self.df.head())
            filename = "error_" + str(id(self.df)) + ".xlsx"
            # self.df.to_excel(filename)
            print_with_info_style("error: requested URL does not fit...")
            time.sleep(1)
            return False
        except Exception as e:
            print_with_failure_style(e)
            pass
        self.make_df_float()
        return self.df
    @property
    def got_items(self):
        if not self.check_api_key_first():
            return False
        self.df_checked = True
        return self.get_items_csv()
    def get_items_csv(self):
        will_be_requested = self.create_url()
        for item in will_be_requested:
            try:
                df = self.convert_csv_df(self.get_csv(item))
            except DfColumnsDoesNotMatch:
                continue
            except Internal:
                continue
            if isinstance(df, pd.DataFrame):
                return df
        return False
    def get_items_json(self):
        """not implemented"""
        # return self.convert_csv_df(self.get_json(self.create_url()))
    def save_excel(self, file_name=None):
        return True
#   ----------------------------------------------------------    / EvdsSorgu
# ------------------------------------------------------------------------------
# /*
#               EvdsSorguSeries
#
# ------------------------------------------------------------------------------
# */
@dataclass
class EvdsSorguSeries(EvdsSorgu):
    """ Seriler iÃƒÂ§in """
    def summary(self):
        if not self.check_api_key_first():
            return False
        self.df = self.got_items
        if not isinstance(self.df, pd.DataFrame):
            return False
        # print(self.df.head())
        return self.df.head()
    def save_excel(self, file_name=None):
        if not self.df_checked:
            self.df = self.got_items
        file_name = self.bfs.xls_name_path
        try:
            os.makedirs(self.bfs.abs_path)
        except:
            pass
        # assert file_name == "ttt"
        if not isinstance(self.df, pd.DataFrame):
            print_with_failure_style("df was not created")
            self.request_or_cache_resultInfo_.excel_saved = False
            return
        self.ExcelSaveClass_ = ExcelSaveClass(self.df, file_name)
        self.ExcelSaveClass_.save_excel__()
        self.request_or_cache_resultInfo_.excel_saved = True
        return True
#   ----------------------------------------------------------    / EvdsSorguSeries
__all__ = [
    'EvdsSorguSeries',
    'EvdsSorgu',
]


# ....................................................................... UrlSeries
class UrlSeries:
    @property
    def domain(self) -> str:
        return "https://evds2.tcmb.gov.tr/service/evds"

    @property
    def alias(self):
        return "series="



# ....................................................................... UrlDataGroup
class UrlDataGroup(UrlSeries):
    @property
    def alias(self):
        return "datagroup="




import warnings
from ..common.common_imports import *
from ..utils.utils_general import *
from ..initial.start_options import *
from ..common.files import Read
from ..common.colors import *
from dataclasses import dataclass 
from abc import ABC 
def get_develop_vers_main():
    parent = Path(__file__).parent
    v = Read(Path(parent / ".." / ".." / "__version__.py"))
    if isinstance(v, str):
        v = v.replace("#", "")
    return v
def version_raw():
    v = get_develop_vers_main()
    return v
def version_display():
    v = get_develop_vers_main()
    import site
    s = site.getsitepackages()
    site_pack = ""
    f = ""
    if len(s) > 1:
        site_pack = s[1]
        f = "from"
    # print_with_success_style(f"evdspy {v} {f} ")
    # print_with_info_style(f"evdspy {v} {f} {site_pack}")
    v = f"evdspy {v} {f} {site_pack}"
    print_with_info_style(v)
@dataclass
class ConfigBase(ABC):
    cancel_request_temp: bool = False
    current_mode_is_test: bool = check_if_this_is_pytest()
    projectName: str = 'evdspy'
    interiorFolder = 'EVDSlocal'
    runtime_file_name_root_path = get_current_dir() / '..'
    runtime_file_name_path = get_current_dir() / '..' / "components" / "options.py"
    user_options_file_name = Path.cwd() / 'IO' / 'options.py'
    user_options_file_name_locked = Path.cwd() / 'IO' / 'options-locked.py'
    version: str = version_raw()
    temp_cancel_mock_request: bool = temp_cancel_mock_request
    def __post_init__(self):
        if self.cancel_request_temp:
            warnings.warn("cancel_request_temp is set to True ")
    def version_display(self):
        version_display()
config = ConfigBase()
__all__ = [
    'ConfigBase',
    'config'
]


import os
from pathlib import Path
from typing import Optional, Any
import pandas as pd
from .github_actions import GithubActions
from ..common.files import Read
from ..config.apikey_class import ApikeyClass

"""Globals  """
EVDS_API_KEY_ENV_NAME = "EVDS_API_KEY"
try:
    import pytest
except Exception:
    pass


def get_api_key():
    import os
    from dotenv import load_dotenv 
    load_dotenv()
    return os.getenv("EVDS_API_KEY")


def key_valid():
    return isinstance(get_api_key(), str) and len(get_api_key()) == 10


def is_df(df: Any):
    return isinstance(df, pd.DataFrame)


def gth_testing():
    return GithubActions().is_testing()


reason_gth = "passing when github Actions "

skip_if_gthub = pytest.mark.skipif(
    gth_testing, reason=reason_gth
)

skip_if_not_keyvalid = pytest.mark.skipif(
    key_valid(), reason='No Api key Valid provided'
)


def get_api_env_key_name():
    return EVDS_API_KEY_ENV_NAME


# import pytest
def get_api_key_file(file_name="api_key.txt", deep=7) -> Optional[Path]:
    def get_file_deep(number):
        folder = Path(".")
        for _ in range(number):
            folder = folder / ".."
            if "pycharmprojects" not in str(folder.absolute()).lower() and  'evdspy-dev' not in str(folder.absolute()).lower() :
                """Do not go back deeper if it is not my folder"""
                return None
            file_namex = folder / file_name
            if file_namex.exists():
                return file_namex
        return None

    return get_file_deep(deep)


def test_get_api_key_file(capsys):
    if GithubActions().is_testing(): return

    with capsys.disabled():
        api_key = get_api_key_file(deep=8)
        print(api_key)


def get_api_key_while_testing():
    file_name = get_api_key_file(deep=7)
    if file_name is None:
        return False
    content = Read(file_name)
    lines = content.splitlines()
    line = tuple(line for line in lines if "evds" in line.lower() )
    api_key = line[0].split("=")[1]
    return str(api_key).strip()


class ApiClassWhileTesting():
    """ApiClassWhileTesting"""

    def __init__(self):
        self.api_key = self.get_api_key()

    def get_api_key(self):
        if GithubActions().is_testing():
            return os.getenv(get_api_env_key_name())
        return get_api_key_while_testing()

    @property
    def key(self):
        return self.api_key

    def __call__(self, *args, **kwargs):
        return self.key


def test_ApiClassWhileTesting(capsys):
    with capsys.disabled():
        api_key = ApiClassWhileTesting().key
        print(ApikeyClass().obscure(api_key))




pop_series_test_content = r"""
#Series_config_file
    E V D S P Y  _  C O N F I G  _  F I L E  ---------------------------------------------
#
# This file will be used by evdspy package (python) in order to help updating
# your series.
# Script will be adding this file when you setup a new project.
# Deleting or modifying its content may require to setup configuration from the beginning
# ----------------------------------------------------------------------------------------
#Begin_series
---Series---------------------------------
filename : t1\series.txt
subject  : visitors
prefix   : EV_
------------SERIES CODES------------------
TP.ODEMGZS.BDTTOPLAM
TP.ODEMGZS.ABD
TP.ODEMGZS.ARJANTIN
TP.ODEMGZS.BREZILYA
TP.ODEMGZS.KANADA
TP.ODEMGZS.KOLOMBIYA
TP.ODEMGZS.MEKSIKA
TP.ODEMGZS.SILI
------------/SERIES CODES------------------
---/Series---------------------------------
--++--
"""


import traceback
from dataclasses import dataclass
from typing import Callable, Any
import pandas as pd

# from ..get_series_indexes import get_series

DF = pd.DataFrame
pos_DF = tuple[DF, ...]


def pass_next_apply(fnc: Callable, *args, **kw):
    result = False
    try:
        result = fnc(*args, **kw)
    except Exception:
        traceback.print_exc()
    return result


@dataclass
class Tables:
    list_: tuple[Any, ...]
    fnc: Callable #= get_series
    cache :bool = True

    def get(self) -> DF:
        data_tuple: tuple[DF, ...] = tuple(map(lambda x: pass_next_apply(self.fnc, x, cache=self.cache), self.list_))
        dt = tuple([x for x in data_tuple if isinstance(x, DF)])
        return self.combine(dt)

    def combine(self, data_tuple: pos_DF) -> DF:
        def fnc(df):
            if 'Tarih' in df.columns:
                df.set_index('Tarih', inplace=True)
            return df

        dfs = [fnc(x) for x in data_tuple]
        combined = pd.concat(list(dfs), axis=1, join='outer')
        return combined

    def __call__(self, *args, **kwargs):
        return self.get()


def get_tables_from_str(string) -> tuple[Any, ...]:
    items = string.splitlines()
    return tuple([x.strip() for x in items if 'bie_' in x])


"""
from ..get_series_indexes import get_series

def tes_tableCombine():
    template = \"""

bie_abreserv
bie_ackap2
bie_akonutsat1
bie_akonutsat2
bie_akonutsat3
bie_akonutsat4
\"""
    tables = get_tables_from_str(template)
    table = Tables(tables , get_series , True )
    df = table()
    assert isinstance(df, DF)
"""




import pandas as pd
from .index_classes import GeneralIndexesDatagroups
from ..common.table import Table2_
from ..components.options_class import SingletonOptions
from ..config.config import ConfigBase
from ..requests_.ev_request import EVRequest
config = ConfigBase()
# https://evds2.tcmb.gov.tr/service/evds/datagroup=bie_yssk&startDate=01-06-2017&endDate=07-09-2030&type=json&key=XXYYZZ
def get_datagroups_with_code(code=1):
    gid = GeneralIndexesDatagroups(code=code, EVRequest_=EVRequest(options_=SingletonOptions()))  #
    json_content = gid.get_json()
    return json_content
def get_keys() -> tuple:
    keys = ('DATAGROUP_CODE', 'DATAGROUP_NAME_ENG',
            'DATAGROUP_NAME', 'FREQUENCY',
            'DATASOURCE_ENG', 'FREQUENCY_STR',
            'START_DATE', 'END_DATE'
            )
    return keys
def get_title(item: dict) -> str:
    return item['DATAGROUP_NAME_ENG']
def get_value_from_datagroups_dict(item: dict, key: str) -> str:
    value = item.get(key, None)  # str(item[key])
    value = str(value)
    return value
import typing as t
def get_titles_and_ids_for_selection(json_content: dict) -> t.List[t.Tuple[str, str]]:
    new_list = []
    for item in json_content:
        title: str = get_value_from_datagroups_dict(item, 'DATAGROUP_NAME_ENG')
        code: str = get_value_from_datagroups_dict(item, 'DATAGROUP_CODE')
        new_list.append((code, title))
    return new_list
def show_datagroups_summary(json_content: dict) -> None:
    keys = get_keys()
    for item in json_content:
        new_list = []
        for key in keys:
            value = get_value_from_datagroups_dict(item, key)
            new_list.append((key, value,))
        title = get_title(item)
        Table2_().show(list_=new_list, title=title, columns=('key', 'value'), skiprow=0)
def json_to_df(json_content: list, code: int) -> None:
    df = pd.DataFrame.from_records(json_content)
    file_name = f"dataGroupOut-{code}.xlsx"
    try:
        df.to_excel(file_name)
    except:
        print(f"could not write excel file {file_name}. =>file is probably open")
        # data_model_type = data_models_dict['json']
    # process_datagroups(data_model_type, json_content)
def get_and_process_datagroups_with_code(code=1):
    json_content: dict = get_datagroups_with_code(code)
    show_datagroups_summary(json_content)
    return get_titles_and_ids_for_selection(json_content)
    # exit()
    # json_to_df(json_content, code)
def get_all_groups():
    # numbers = (1, 2, 3, 4, 5, 12, 13, 14, 15, 18, 19, 20, 21, 22, 23, 6, 7, 24, 25, 26, 27, 28, 95,)
    numbers = (18, 19,)
    for num in numbers:
        try:
            get_and_process_datagroups_with_code(num)
        except Exception as exc:
            print(exc)
            pass



# ...........................................................................................
import traceback
import requests
from typing import Any


# ...........................................................................................
from evdspy.EVDSlocal.config.apikey_class import ApikeyClass
from evdspy.EVDSlocal.index_requests.user_requests.Url_builder import UrlBuilder 
from evdspy.EVDSlocal.index_requests.user_requests.Proxy_manager import ProxyManager 
from evdspy.EVDSlocal.index_requests.user_requests.Request_config import RequestConfig
from evdspy.EVDSlocal.index_requests.user_requests.User_request_utils import cache_or_raw_fnc
from evdspy.EVDSlocal.requests_.real_requests import RealRequestWithParam
from evdspy.EVDSlocal.utils.utils_test import ApiClassWhileTesting
from evdspy.EVDSlocal.utils.github_actions import PytestTesting, GithubActions

# ...........................................................................................


# ....................................................................... ApiRequester
class ApiRequester:
    def __init__(self, url_builder: UrlBuilder, proxy_manager: ProxyManager):
        self.url_builder = url_builder
        self.proxy_manager = proxy_manager
        self.proxies = self.proxy_manager.get_proxies()
        self.url = self.url_builder.url
        self.response = None

    def get(self):
        return self.request()

    def __call__(self, *args, **kwargs):
        return self.get()

    def dry_request(self) -> RequestConfig:
        api_key = self.get_api_key(check=False)
        print(f"""
---------------------------------
    [debug mode is turned on]
        api_key = {self.obscure(api_key)}
        proxies = {self.proxies}
        url  = {self.url}
    ! request was not made because debug mode is turned on
    in order to make the request run get_series(... , debug = False )
---------------------------------
""")
        return self.url_builder.config

    def is_response_ok(self, response):
        return isinstance(response, requests.Response) \
            and response.status_code == 200

    @staticmethod
    def obscure(string: str):
        return ApikeyClass().obscure(string)

    def get_api_key(self, check=True) -> str:
        if PytestTesting().is_testing() or GithubActions().is_testing():
            api_key = self.get_api_key_while_testing()
            if not api_key:
                api_key = ApikeyClass().get_valid_api_key(check=check) 
        else:
            api_key = ApikeyClass().get_valid_api_key(check=check)
        return api_key

    def get_api_key_while_testing(self):
        return ApiClassWhileTesting()()

    def request(self) -> Any:
        api_key = self.get_api_key()
        if api_key is False:
            if GithubActions().is_testing():
                return self.dry_request()
            if PytestTesting().is_testing():
                raise NotImplementedError
        proxies = self.proxy_manager.get_proxies()

        def local_request(url: str) -> requests.Response:
            requester = RealRequestWithParam(url,
                                             proxies=proxies,
                                             api_key=api_key)
            return requester.request()

        request_func = cache_or_raw_fnc(local_request,
                                        cache=self.url_builder.config.cache)
        response = False
        try:
            response = request_func(self.url)
        except Exception:
            traceback.print_exc()
        self.response = response
        if not self.is_response_ok(response):
            return False
            # raise HTTPError(response=response)
        return response.json()

    def is_ok(self) -> bool:
        response = self.response
        ok = isinstance(response, requests.Response) and response.status_code == 200
        if not ok:
            self.dry_request()
            raise ValueError("request Not Ok")
        else:
            print("<data is here>")
        return ok




from .index_classes import GeneralIndexesCategories
from ..components.options_class import load_options
from ..requests_.ev_request import EVRequest
def get_categories() -> str:
    """get_categories"""
    # EVRequest_: EVRequest = EVRequest()
    return GeneralIndexesCategories(EVRequest_=EVRequest(options_=load_options())).create_url()
def get_categories_data() -> str:
    """get_categories_data"""
    csv = GeneralIndexesCategories(EVRequest_=EVRequest(options_=load_options())).get_csv()
    # Write("categories_list.txt", csv)
    return csv
def get_table_header():
    from rich.table import Table
    table = Table(title="Categories / Kategoriler")
    table.add_column("ID/Sayi", justify="right", style="cyan", no_wrap=True)
    table.add_column("Title/Baslik (ENG)", style="magenta")
    table.add_column("Title/Baslik (TR)", justify="right", style="green")
    return table
from typing import List
import typing as t
def get_category_name(code: t.Union[int, str] = 1) -> t.Union[str, None]:
    if isinstance(code, str):
        code = int(code)
    csv_buffer: str = get_categories_data()
    lines = csv_buffer.splitlines()
    list_of_categs = {}
    for item in lines[1:]:
        id, cat_eng, cat_tr = split_wisely(item)
        list_of_categs.update({int(id): cat_tr})
    return list_of_categs.get(code, None)
def display_categories() -> t.Union[List[tuple], None]:
    """display_categories"""
    from rich.console import Console
    csv_buffer: str = get_categories_data()
    table = get_table_header()
    lines = csv_buffer.splitlines()
    list_of_categs = []
    if not lines:
        return
    for item in lines[1:]:
        id, cat_eng, cat_tr = split_wisely(item)
        list_of_categs.append((id, cat_eng, cat_tr,))
        table.add_row(id, cat_eng, cat_tr)
    console = Console()
    console.print(table)
    return list_of_categs
def split_wisely(line: str) -> tuple:
    """split_wisely"""
    parts = line.split(",")
    id, cat_eng, cat_tr = "", "", ""
    if len(parts) == 3:
        id, cat_eng, cat_tr = parts
    if len(parts) == 4:
        id, cat_eng, cat_tr, cat_tr2 = parts
        cat_tr = f"{cat_tr}, {cat_tr2}"
    if id:  # '1.0'
        id = id.split(".")[0]
    return id, cat_eng, cat_tr


# ....................................................................... DataProcessor
import traceback
from typing import Any

import pandas as pd

from evdspy.EVDSlocal.index_requests.index_util_funcs import make_df_float, json_to_df
from evdspy.EVDSlocal.index_requests.user_requests.User_req_typings import T_maybeDf


class DataProcessor:
    def __init__(self, data: Any):
        self.data = data

    def process_to_dataframe(self) -> T_maybeDf:
        if self.data is False:
            return False
        try:
            df = json_to_df(self.data)
        except Exception as e:
            print(e)
            traceback.print_exc()
            return None
        if isinstance(df, pd.DataFrame):
            df = make_df_float(df)
        return df

    def __call__(self, *args, **kwargs) -> T_maybeDf:
        return self.process_to_dataframe()


def test_DataProcessor(capsys):
    with capsys.disabled():
        d = DataProcessor(False)
        print(d)



from .Url_builder import UrlBuilder
from .Data_processor import DataProcessor
from .Proxy_manager import ProxyManager
from .Serialize import Serialize
from .Url_series import UrlSeries, UrlDataGroup
from .User_req_typings import *
from .Api_requester import ApiRequester
from .Request_config import RequestConfig



import pandas as pd
import typing as t
from ..common.colors import print_with_failure_style
from ..components.excel_class import correct_folder


def make_date_first_column_helper(df, col_name):
    if col_name not in df.columns:
        return df
    cols = [col_name] + [col for col in df if col != col_name]
    df = df[cols]
    return df


def make_date_first_column(df):
    col_names = ('YEARWEEK', 'Tarih',)
    for col_name in col_names:
        df = make_date_first_column_helper(df, col_name)
    return df


def drop_unix(df):
    UnixTime = 'UNIXTIME'
    if UnixTime in df.columns:
        df.drop(UnixTime, axis=1, inplace=True)
    return df


def json_to_df(json_content: t.Union[list, dict]):
    if hasattr(json_content, 'items'):
        json_content = json_content['items']
    df = pd.DataFrame.from_records(json_content)
    df = drop_unix(df)
    df = make_date_first_column(df)
    return df


def make_df_float(df):
    def check_cols(x: str):
        items = (
            "Tarih" not in x,
            "WEEK" not in x,
            "DAY" not in x,
            "YEAR" not in x,
        )
        return all(items)

    columns = (x for x in df.columns if check_cols(x))
    other_cols = (x for x in df.columns if not check_cols(x))
    old_df = df
    try:
        df = df[columns].astype(float)
    except:
        # print_with_failure_style("Could not convert some columns to float type...")
        ...
    for other_column in other_cols:
        df[other_column] = old_df[other_column]
    return df


def json_to_excel(json_content: list, file_name, try_float=True):
    file_name = correct_folder(file_name, "xlsx")
    df = json_to_df(json_content)
    if try_float:
        df = make_df_float(df)
    try:
        df = make_date_first_column(df)
        df.to_excel(file_name)
    except:
        print(f"could not write excel file {file_name}. =>file is probably open")



import json
from abc import ABC


# ....................................................................... Serialize
class Serialize(ABC):
    """To check whether two Config Requests are perfect substitutes """

    def serialize(self):
        return json.dumps(
            self,
            default=lambda o: o.__dict__,
            sort_keys=True,
            indent=4)

    @property
    def hash(self) -> str:
        import hashlib
        return str(int(hashlib.sha256(self.serialize().encode('utf-8')).hexdigest(), 16) % 10 ** 8)

    def __eq__(self, other):
        return self.hash == other.hash





from ..common.common_imports import *
from dataclasses import dataclass, field
from ..config.config import *
from evdspy.EVDSlocal.console.screen import *
from ..common.prog import rich_sim
from ..common.colors import *
# from evdspy.utils.utils_general import *
indent = " " * 15

from typing import Union, List, Callable
do_nothing = lambda: True
@dataclass
class MenuItem:
    """
    MenuItem
    """
    func: Union[Callable, str]
    disp: str
    sleeptime: int = 0
    clear: bool = False
    name: str = ""
    callable_func: Callable = do_nothing
    def __post_init__(self):
        if callable(self.func):
            self.name = self.func.__name__
            self.callable_func = self.func
        elif isinstance(self.func, str):
            self.name = self.func
        else:
            self.name = self.disp
        self.get_sleep_times()
        self.get_screen_clear_bool()
    def get_sleep_times(self):
        sleeptime_default = 1
        obj = {"check": sleeptime_default, "setup": sleeptime_default, "get": sleeptime_default,
               "get_categories_main": sleeptime_default}
        self.sleeptime = obj.get(self.name, sleeptime_default)
    def get_screen_clear_bool(self):
        obj = {"check": False, "get": False, "help_": False, "get_categories_main": False}
        self.clear = obj.get(self.name, True)
@dataclass
class MenuMaker:
    menu_items: field(default_factory=List[MenuItem])
    # menu_items_display : field(default_factory=list)
    message: str = "\n" + " " * 25 + "Selection ? "
    exit_item: bool = True
    exit_: bool = False
    exit_menu_call_back: Callable = do_nothing
    def __post_init__(self):
        self.add_exit_item()
        Screen().clear()
    def exit(self):
        self.exit_ = True
        self.exit_menu_call_back()
        return
        # self.display()
    def menu_header(self):
        width = 60
        def make_indent(element, num=width):
            indent = " " * (num - len(element))
            return f"{indent}{element}"
        def make_center(element, num=width):
            left = " " * round((num - len(element)) / 2)
            return f"{left}{element}"
        menu_title = make_center("M E N U")
        reminder = "to get the latest version:   pip install -U evdspy"
        reminder = make_indent(reminder)
        version_line = make_indent(config.version)
        logo = make_indent("evdspy @2022")
        line_1 = "-" * width
        print_with_info_style(reminder)
        header = f"""
{line_1}
{version_line}
{line_1}
{logo}
{menu_title}
{line_1}
"""
        return header
    def title(self):
        print("\n" * 2)
        print(self.menu_header())
    def add_exit_item(self):
        if self.exit_item:
            exit_menu_item = MenuItem(self.exit, "EXIT (console)")
            self.menu_items = tuple(list(self.menu_items) + [exit_menu_item])
    def make_table(self, items):
        from rich.console import Console
        from rich.table import Table
        def get_color(item: str):
            font_color = "green"
            if "false" in item.lower():
                font_color = "red"
            return font_color
        table = Table(title="M E N U ")
        table.add_column("Select", justify="right", style="cyan", no_wrap=True)
        table.add_column("Job / Request ", style="green")
        table.add_column("Explanation ", justify="right", style="magenta")
        for index, item in enumerate(items):
            font_color = get_color(item.disp)
            table.add_row(f"{index + 1}", str(item.disp).upper(), "", style=font_color)
        console = Console()
        console.print(table)
    def display(self):
        if self.exit_:
            return
        self.make_table(self.menu_items)
        self.get_choice()
    def display_old(self):
        if self.exit_:
            return
        self.title()
        for index, item in enumerate(self.menu_items):
            dec = f"{indent}{index + 1}. {item.disp} "
            print_menu_item_style(dec)
        self.get_choice()
    def menu_exists_and_callable(self, a: str):
        if str(a).isnumeric() and int(a) in (x for x in range(1, len(self.menu_items) + 1)):
            menu_item: MenuItem = self.menu_items[int(a) - 1]
            func = menu_item.func
            return callable(func), menu_item.callable_func, menu_item
        return False, False, False
    def get_choice(self):
        if self.exit_:
            return
        if config.current_mode_is_test:
            return
        ans = input(self.message)
        cond, func, menu_item = self.menu_exists_and_callable(ans)
        if cond:
            wait_seconds = menu_item.sleeptime
            if menu_item.clear:
                Screen().clear()
            v = func()
            if v == -1:
                self.exit_ = True
            rich_sim(wait_seconds, "completing")
        else:
            Screen().clear()
        return self.display()
__all__ = [
    'MenuItem',
    'MenuMaker',
]
def test_show():
    print("test")
def test_show2():
    print("test 2")
def test_show3():
    print("test 3")
def test_f():
    t1 = MenuItem(test_show, "test_show")
    t2 = MenuItem(test_show2, "test_show2")
    t3 = MenuItem(test_show3, "test_show3")
    m = MenuMaker(menu_items=[t1, t2, t3])
    m.display()



from ..common.common_imports import *
import os
opsys_dict = {
    'nt': 'windows',
    'posix': 'linux/mac',
    'default': 'linux/mac'
}
opsys_name = opsys_dict.get(os.name, opsys_dict.get('default', None))
clear_dict = {
    'windows': 'cls',
    'linux/mac': 'clear',
    'default': 'clear'
}
clear_command_name = clear_dict.get(opsys_name, clear_dict.get('default', None))
class Screen:
    def clear(self):
        os.system(clear_command_name)



# functions that will be loaded
import platform
import sys
import time
import pandas as pd
from evdspy.EVDSlocal.console.menu import MenuMaker, MenuItem
from .cmd_line_classes import CommandLineCommandClass
from evdspy.EVDSlocal.initial_setup.api_key_save import save_api_key_to_file
from typing import Callable
from ..common.prog import rich_sim
from ..common.table import Table2_
from ..index_requests.categories import get_category_name
from ..index_requests.datagroup_individual import get_datagroup_individual_with_code, get_series_list_of_subject, \
    get_df_datagroup
from ..messages.error_messages import *
from ..messages.error_classes import *
from evdspy.EVDSlocal.initial.load_modules import LoadModulesClass
from ..initial_setup.initial_setups import *
from ..requests_.my_cache import delete_cache_folder
from ..setup_project.user_setup import start_setup_series, start_setup_config
from evdspy.EVDSlocal.initial_setup.setup_folders import check_folders_setup_necessary, check_setup
from ..messages.help_messages import welcome_message
from ..utils.utils_general import *
from ..index_requests.datagroups import get_datagroups_with_code
assert get_datagroups_with_code
assert get_category_name
assert get_series_list_of_subject
"""
Loaded functions for DF
get_df_datagroup
"""
assert callable(get_df_datagroup)
def check_which_command_params(params, extra_params: dict):
    create_series_cmd = CommandLineCommandClass(name_list=("create", "series",),
                                                func=setup_series)
    help_cmd = CommandLineCommandClass(name_list=("help",),
                                       func=help_evds)
    create_options_cmd = CommandLineCommandClass(name_list=("create", "options",),
                                                 func=create_options_file)
    get_cmd = CommandLineCommandClass(name_list=("get",), func=get)
    setup_cmd = CommandLineCommandClass(name_list=("setup",), func=setup)
    menu_cmd = CommandLineCommandClass(name_list=("menu",), func=menu)
    save_cmd = CommandLineCommandClass(name_list=("save",), func=save)
    console_cmd = CommandLineCommandClass(name_list=("console",), func=console_load)
    menu_list = (create_series_cmd, create_options_cmd, get_cmd, setup_cmd, menu_cmd, help_cmd, save_cmd, console_cmd)
    def check_item_exists(element, format_cmd_list):
        return element in format_cmd_list
    def check_item_exists_wrapper(item):
        return check_item_exists(item, params)
    def check_cmd_will_apply(cmd_class: CommandLineCommandClass):
        # return check_item_exists, items
        return all(map(check_item_exists_wrapper, cmd_class.name_list))
    from itertools import compress
    functions_will_apply_iter = tuple(map(check_cmd_will_apply, menu_list))
    functions_will_apply = list(compress(menu_list, functions_will_apply_iter))
    for fnc_class in functions_will_apply:
        if callable(fnc_class.func):
            fnc_class.func()
    return functions_will_apply
def menu():
    """
    Displays user-friendly menu screen for many functionalities
    such as saving API key, getting data, categories etc.
    :return: None
    """
    menu_helper()
# -------------------------------- Entry point ---cmd line prompt------------------------------------
def console_main(test_args=None):
    # test_args = tuple(["series", "create"])
    if test_args:
        params_command = test_args
    else:
        params_command = sys.argv[1:]
    params_command_as_dict = arg_acc(params_command)
    if not params_command:
        welcome_message()
    check_which_command_params(params_command, arg_acc())
# -------------------------------- Entry point ---cmd line prompt------------------------------------
def console_main_from_the_menu():
    welcome_message()
def easy_setup():
    setup()
def setup_again():
    setup()
def setup():
    SetupInitial().setup()
    start_setup_config(onsetup=True)
    create_series_text_example(onsetup=True)
def help_evds():
    display_help_messages()
def help_():
    return help_evds()
def h():
    return help_evds()
def help_evdspy():
    return help_evds()
def check():
    print_with_updating_style("checking....")
    lmc = LoadModulesClass()
    lmc.check()
def show():
    lmc = LoadModulesClass()
    lmc.series_from_file()
    lmc.summary()
def get_input(msg):
    if config.current_mode_is_test:
        return "test"
    print_get_input_style(msg)
    ans = input()
    return str(ans)
def wait(num: int):
    time.sleep(num)
def save_api_key_to_file_main(api_key):
    @dataclass
    class transfer():
        get_input: Callable = get_input
        wait: Callable = wait
    save_api_key_to_file(instance=transfer(), api_key_answer=api_key)
def console_load():
    ...
def check_setup_runtime():
    return not check_folders_setup_necessary()
def check_if_api_set_before_get():
    if not check_api_key():
        print_with_failure_style(api_key_not_set_msg)
        wait(2)
        if get_input("yes/no ?  ").lower() not in ["yes", "y"]:
            wait(2)
            return
        # below the function checks
        # if api key is valid then saves it
        if not set_apikey_input():
            return False
        print_with_success_style("finished")
        return True
    return True
def get():
    if not check_setup():
        print_with_updating_style("Checking setup...")
        # setup_folders()
        SetupInitial().setup()
    if not check_if_api_set_before_get():
        return
    try:
        lmc = LoadModulesClass()
        lmc.series_from_file()
        # lmc.display_items(lmc.series_from_file())
        lmc.summary()
        # try:
        #     lmc.evds_list[0].df.plot(x='TP.ODEMGZS.BDTTOPLAM', y='Index', style='o')
        # except:
        #     pass
    except SeriesFileDoesNotExists:
        print("...")
        # print(SeriesFileDoesNotExists().message)
def get_df_test():
    LoadModulesClass().series_from_file()
    LoadModulesClass().summary()
    return LoadModulesClass().evds_list[0].df
from ..config.apikey_class import *
from evdspy.EVDSlocal.requests_.lower_energy import *
from typing import List
from evdspy.EVDSlocal.index_requests.datagroups import get_and_process_datagroups_with_code, \
    get_all_groups, get_datagroups_with_code
from evdspy.EVDSlocal.index_requests.index_util_funcs import json_to_df
import typing as t
from functools import partial
def get_input_for_categs_data_groups(list_of_categs: List[tuple], display_fn, callback: t.Callable = None):
    callback_forward = partial(get_input_for_categs_data_groups, list_of_categs, display_fn, None)
    callables = (display_fn, callback_forward)
    if config.current_mode_is_test:
        return
    ids = (str(x[0]) for x in list_of_categs)
    # print_with_success_style('Preparing the menu...')
    # print("\n\n")
    # time.sleep(2)
    ans = input("Choose a Category number [e/exit to go back] Selection Number==>")
    if str(ans).lower().strip() in ('e', 'exit'):
        if callable(callback):
            print("now calling back ", display_fn, callback)
            display_fn()
            callback()
        return
    if str(ans) in ids:
        codes_and_titles = get_and_process_datagroups_with_code(int(ans))
        get_input_for_categs_data_groups_deeper(codes_and_titles, callables)
def get_input_for_categs_data_groups_deeper(codes_and_titles: List[tuple], callables: t.Tuple[any, partial] = None):
    codes = tuple(x[0] for x in codes_and_titles)
    nums = tuple(str(x + 1) for x in range(len(codes_and_titles)))
    titles = tuple(x[1] for x in codes_and_titles)
    combine_list = list(zip(nums, codes, titles))
    if config.current_mode_is_test:
        return
    # print(codes_and_titles)
    # print("\n\n")
    Table2_().show(list_=combine_list, title='', columns=('Selection Number', 'code', 'title'), skiprow=0)
    # print_with_success_style('Preparing the menu...')
    # time.sleep(2)
    ans = input("Choose a Category number  [e/exit to go back]  Selection Number==>")
    if str(ans).lower().strip() in ('e', 'exit'):
        if callables:
            for callback in callables:
                if callable(callback) or isinstance(callback, partial):
                    callback()
                    time.sleep(1)
        return
    if str(ans) in nums:
        code_str = codes[int(ans) - 1]
        try:
            json_content = get_datagroup_individual_with_code(code_str)
            # print(json_content)
            df = json_to_df(json_content)
            print(df.head())
            series_docs = get_series_list_of_subject(code_str)
            df_exp: pd.DataFrame = json_to_df(series_docs)
            print_with_success_style(
                    "Getting explanations for the series...after this process both excel files will be created...[1-Data , 2-Explanations] "
            )
            time.sleep(1)
            print(df_exp.head())
            # print(series_docs)
        except:
            print_with_failure_style(
                    f"...json  content is not proper to convert to a pd.DataFrame .. passing this one... code : {code_str}")
            pass
        df_exp: pd.DataFrame = json_to_df(series_docs)
        print_with_success_style(
                "Getting explanations for the series...after this process both excel files will be created...[1-Data , 2-Explanations] "
        )
        time.sleep(1)
        print(df_exp.head())
        # print(series_docs)
    print(ans, " ... done.")
    get_input_for_categs_data_groups_deeper(codes_and_titles, callables)
    # codes_and_titles = get_and_process_datagroups_with_code(int(ans))
    # get_input_for_categs_data_groups_deeper(codes_and_titles)
def get_categories_main():
    if not check_if_api_set_before_get():
        return
    # from ..index_requests.categories import get_categories_data
    from ..index_requests.categories import display_categories
    list_of_categs: List[tuple] = display_categories()  # [( id , eng , tr ) , ( id , eng , tr ) ...]
    get_input_for_categs_data_groups(list_of_categs, display_fn=display_categories)
    # categories = get_categories_data()
    # inspect(categories, all=True)
    # print_with_success_style(categories)
def get_datagroups_data_main():
    ...
    # from ..index_requests.datagroups import get_datagroups_data
    # categories = get_datagroups_data()
    # inspect(categories, all=True)
    # print_with_success_style(categories)
def check_api_key_with_low_energy(api_key_from_user):
    # check api key with cache
    # send => basic_for_test > performance.lower_energy
    if not apikey_works(api_key_from_user):
        print_with_failure_style("Api key did not work! ")
        return False
    print_with_success_style("Api key works")
    return True
def obvious_checks_of_api_key(api_key: str) -> bool:
    conds: tuple = (isinstance(api_key, str), not len(api_key) < 10,)
    return all(conds)
def save_apikey(api_key: str = None):
    if api_key is None or not isinstance(api_key, str):
        return set_apikey_input()
    check_apikey_and_then_save(api_key)
def check_apikey_and_then_save(api_key_from_user):
    if not obvious_checks_of_api_key(api_key_from_user):
        print_with_failure_style(
                f"The text you entered does not meet criteria to be an valid api key. \n"
                f"{indent}(length of api key cannot be less then 10)  ")
        return False
    """this function will save redundant requests by catching earlier check results """
    if not check_api_key_with_low_energy(api_key_from_user):
        return False
    """"""
    """ SAVE IT TO FILE """
    save_api_key_to_file_main(api_key_from_user)
    """ SAVE IT FOR RUNTIME """
    ApikeyClass().set_api_key_runtime(value=api_key_from_user)
    # show_apikey()
    return True
def set_apikey_input(api_key=None, from_console=False):
    exit_words = ["e", "exit"]
    if config.current_mode_is_test:
        print_with_updating_style("pytest returns ")
        ans = "test"
    else:
        if not from_console:
            ans = get_input("Api key [ `e` or `exit` to cancel] : ")
        else:
            ans = api_key
    api_key_from_user = str(ans)
    if api_key_from_user.lower() in exit_words:
        print_with_updating_style("exited")
        return False
    return check_apikey_and_then_save(api_key_from_user)
from evdspy.EVDSlocal.initial_setup.api_key_save import check_api_key_on_load
def check_api_key():
    check_api_key_on_load()
    apikeyobj = ApikeyClass()
    return apikeyobj.get_valid_api_key()
def show_apikey():
    apikeyobj = ApikeyClass()
    # k = apikeyobj.get_api_keys_dict()
    valid = apikeyobj.get_valid_api_key()
    disp = f"""
    valid api key : {valid}
"""
    print_with_success_style(disp)
from ..series_format.series_creator import *
def create_series_file():
    return create_series_text_example()
def csf():
    return create_series_file()
def setup_series():
    start_setup_series()
def setup_user_options():
    start_setup_config(onsetup=False)
def create_options_file():
    setup_user_options()
def cof():
    create_options_file()
def setup_series_steps():
    return setup_series()
def console():
    return -1
def get_develop_vers_main():
    # print_with_updating_style(str(Path.cwd()))
    parent = Path(__file__).parent
    v = Read(Path(parent / ".." / ".." / "__version__.py"))  # EVDSlocal/ version.py
    return v
def version():
    from evdspy.EVDSlocal.console.menu_logo_header import version_logo
    logo_version, reminder = version_logo()
    print_with_success_style(logo_version)
    print_with_info_style(reminder)
def py_version():
    #
    # print_with_creating_style(sys.version_info)
    # print_with_creating_style(python_version())
    print_with_creating_style(sys.version)
    # print_with_creating_style(str(platform.python_version_tuple()))
def check_compat():
    v_tuple = platform.python_version_tuple()
    # v_tuple = "3.6.0".split(".")
    v_tuple = tuple(map(lambda x: int(x), v_tuple))
    v = sys.version  # sys.version_info
    if (3, 11, -1) < v_tuple:
        print_with_failure_style(
                f"Your python version is {v}. This program may break because it is currently only compatible with versions between 3.7 and 3.11")
    elif (3, 7, 0) > v_tuple:
        print_with_failure_style(
                f"Your python version is {v}. This program may break because it is currently only compatible with versions between 3.7 and 3.11")
    else:
        print_with_success_style(
                f"Your python version is {v} This program was tested with this version and runs properly. However, "
                f"if you notice a bug or if your version breaks at runtime please feel free to open a PR on github.")
def save(*args):
    return save_apikey(*args)
def remove_cache():
    # TODO activate
    msg = "This will attempt to delete cache files in order to make new requests. \n " \
          "If you created your options file you may choose caching period to a more frequent one. \n" \
          "this function was added to force making new requests ignoring recent ones. \n" \
          "Proceed? (Y/n)"
    ans = get_input(msg)
    if ans not in ["y", "Y"]:
        print_with_failure_style("exited without removing the cache")
        return
    print_with_success_style("removing cache...")
    if delete_cache_folder():
        rich_sim(3, "processing...")
        print_with_success_style("cahce folder removed...")
    else:
        print_with_failure_style("cahce folder coould not be removed...")

def main_exit_function():
    ...
def menu_display():
    funcs = [
            ("check setup", check),
            ("setup", setup),
            ("create user options file", create_options_file),
            ("create series file", create_series_file),
            ("add new series group ", setup_series_steps),
            ("get data", get),
            ("get categories (get all series of a datagroup)", get_categories_main),  # next version
            # ("get data groups (on development) ", get_datagroups_data_main),  # next version
            ("help", help_),
            ("show api key", show_apikey),
            ("save api key to file", set_apikey_input),
            ("remove cache folders", remove_cache),
            ("evdspy as a command line prompt", console_main_from_the_menu),
            ("version", version),
            ("py version", py_version),
            ("check compatibility of your python version", check_compat),
    ]

    menu_items = list(map(lambda x: MenuItem(x[1], x[0]), funcs))
    MenuMaker(
            menu_items=menu_items,
            exit_item=True,
            exit_menu_call_back=main_exit_function
    ).display()
def menu_helper():
    # TODO config
    global menu_already_displayed
    menu_already_displayed = True
    menu_display()




def menu_onload():
    if CurrentState().menu_will_open:
        menu_helper()
def show_menu():
    menu_helper()
# get_df_datagroup = get_df_datagroup



from dataclasses import dataclass, field
from typing import List, Tuple, Union
from evdspy.EVDSlocal.components.api_params import UrlParam, convert_list_params_to_str, EVDSApiDomain, Frequency, \
    get_enum_with_value, FrequencyEnum, FormulasEnum, Formulas, AggregationEnum, Aggregations
from evdspy.EVDSlocal.components.api_params import *
import typing as t
from enum import Enum
from .api_params_enums import get_enum_with_value
class ApiParamsNotExpectedFormat(BaseException):
    """ApiParamsNotExpectedFormat"""
@dataclass
class URLClass:
    url_items: Union[List[UrlParam], Tuple[UrlParam]] = field(default_factory=list)
    url: str = None
    url_only_required = None
    url_items_required = None
    url_report: str = ""
    url_explanation: str = ""
    domain: str = str(EVDSApiDomain())
    def __post_init__(self):
        self.url_items_required = tuple(x for x in self.url_items if x.required)
        self.refresh_url()
    def safe_to_report_url(self, url_parts=None):
        if url_parts is None:
            url_parts = self.url_items
        url_parts = [x for x in url_parts if not x.hidden]
        url: str = str(self.domain) + convert_list_params_to_str(url_parts, str)
        return url
    def create_report(self, url_parts=None):
        if url_parts is None:
            url_parts = self.url_items
        self.url_report = convert_list_params_to_str(url_parts, "report")
        self.url_explanation = convert_list_params_to_str(url_parts, "explanation")
        return [self.url_report, self.url_explanation]
    def create_url(self, url_parts: Union[List, Tuple] = None):
        """both for required and complete """
        if url_parts is None:
            url_parts = self.url_items
        url: str = str(self.domain) + convert_list_params_to_str(url_parts, str)
        return url
    def refresh_required(self):
        self.url_items_required = tuple(x for x in self.url_items if x.required)
    def refresh_url(self):
        self.create_url_complete()
        self.create_url_required()
        # self.create_report()
    def create_url_complete(self):
        self.url: str = self.create_url(self.url_items)
    def create_url_required(self):
        self.url_only_required: str = self.create_url(self.url_items_required)
    def add_item(self, item: UrlParam) -> None:
        self.url_items = list(self.url_items) + [item]
        self.refresh_required()
        self.refresh_url()
    def add_frame(self, code_str: str):
        self.add_item(DataGroup(code_str))
        self.refresh_url()
    def add_apikey(self, api_key=None):
        """ Not needed because apikey was moved to header """
        return
        # if api_key is None:
        #     api_key = ApikeyClass().get_valid_api_key()
        # self.add_item(Apikey(api_key))
        # self.refresh_url()
    def add_mode(self, mode_enum: t.Union[Enum, str, int] = 1):
        mode_enum = get_enum_with_value(mode_enum, ModeEnumDatagroup, ModeEnumDatagroup.all_groups)
        self.add_item(ModeParamDataGroups(mode_enum))
        self.refresh_url()
    def add_code(self, code: t.Union[int, str]):
        self.add_item(CodeParamDataGroups(code))
        self.refresh_url()
    def add_frequency(self, frequency: Union[int, str]):
        if isinstance(frequency, str):
            frequency = int(frequency)
        self.add_item(Frequency(get_enum_with_value(frequency, FrequencyEnum)))
        self.refresh_url()
    def add_formulas(self, formulas: Union[int, str], number_of_series: int):
        if isinstance(formulas, str):
            formulas = int(formulas)
        self.add_item(Formulas(get_enum_with_value(formulas, FormulasEnum), number_of_repeat=number_of_series))
        self.refresh_url()
    def add_aggtype(self, aggtype: str, number_of_series: int):
        self.add_item(Aggregations(get_enum_with_value(aggtype, AggregationEnum), number_of_repeat=number_of_series))
        self.refresh_url()



import typing as t
from enum import Enum, auto
# ------------------------------------------------------- E N U M S -------------------------------------------------------
class EnumBase(Enum):
    ...
# --------------------------------FrequencyEnum---------------------
class FrequencyEnum(Enum):
    daily = 1
    business_day = 2
    weekly = 3
    semimonthly = 4
    monthly = 5
    quarterly = 6
    semi_annual = 7
    annual = 8
    default = 5
# --------------------------------DomainParts---------------------
class DomainParts(Enum):
    categories = auto()
    datagroup = auto()
    series = auto()
# --------------------------------AggregationEnum---------------------
class AggregationEnum(Enum):
    average = "avg"
    min = "min"
    max = "max"
    first = "first"
    last = "last"
    cumulative = "sum"
    default = "avg"
# --------------------------------ModeEnumDatagroup---------------------
class ModeEnumDatagroup(Enum):
    all_groups = 0
    data_group = 1
    subject_group = 2
# --------------------------------CodeEnumDatagroup---------------------
# class CodeEnumDatagroup(Enum):
#     code_int = 1
#
#     data_group = 1
#     subject_group = 2
# --------------------------------Separator---------------------
class Separator(Enum):
    period = "."
    comma = ","
# --------------------------------dataTypeEnum---------------------
class dataTypeEnum(Enum):
    csv: str = "csv"
    json_type: str = "json"
# --------------------------------FormulasEnum---------------------
class FormulasEnum(Enum):
    level = 0
    percentage_change = 1
    difference = 2
    yoy_percentage_change = 3
    yoy_difference = 4
    per_cha_com_end_of_pre_yea = 5
    dif_com_end_of_pre_yea = 6
    moving_average = 7
    moving_sum = 8
    default = 0
    # ------------------------------------------------------- get_enum_with_value -------------------------------------------------------
    """ typical usages
        get_enum_with_value( 5 , Frequency , Frequency.Monthly )
    """
"""
        key: t.Union[Enum, str, int],
        # enum_: t.Union[str, int, t.Callable , Enum ],
        enum_: Enum,
        default_value: t.Union[str, int, t.Callable] = None
                key: t.Union[Enum, str, int],
        # enum_: t.Union[str, int, t.Callable , Enum ],
        enum_: any,
        default_value: any = None
"""
def get_enum_with_value(
        key: t.Union[Enum, str, int],
        enum_: t.Union[str, int, t.Callable , Enum ],
        # enum_: Enum,
        default_value: t.Union[str, int, t.Callable , Enum ] = None
) -> Enum :
    """
    :rtype: object
    """
    if isinstance(key, Enum):
        return key
    if isinstance(key, int):
        key = str(key)
    if hasattr(default_value, "value"):
        """otherwise probably None """
        default_value = getattr(default_value, "value", None)
    dd = {str(x.value): x for x in enum_}
    enum_value: Enum = dd.get(key, None)
    if hasattr(enum_value, "value"):
        v = getattr(enum_value, "value")
        return v
    if not default_value:
        return getattr(default_value, "default")
    return default_value
# ------------------------------------------------------- / get_enum_with_value -------------------------------------------------------


import os

def get_proxies_env_helper(a  , b  , a1 , b1 ) -> dict:
    def get_one(x, y):
        if x is None:
            return y
        return x

    a = get_one(a, a1)
    b = get_one(b, b1)

    if a is None and b is None:
        return False
    if a and not b:
        return {"http": a}
    if b and not a:
        return {"https": b}
    return {
        "http": a,
        "https": b
    }
def get_proxies_env() -> dict:
    from ..config.dotenv import load_env
    load_env()  # http_proxy
    a = os.getenv("http_proxy")
    a1 = os.getenv("PROXY_http")
    b = os.getenv("https_proxy")
    b1 = os.getenv("PROXY_https")
    return get_proxies_env_helper( a , b, a1 , b1 )





from dataclasses import dataclass, field
from typing import Union, Any, Optional, Literal, Callable
import pandas as pd
from evdspy.EVDSlocal.index_requests.user_requests.Request_config import RequestConfig
from evdspy.EVDSlocal.index_requests.get_series_indexes_utils import (
    default_start_date_fnc,
    default_end_date_fnc,
)
from evdspy.EVDSlocal.index_requests.get_series_indexes import get_series

def initial_api_process_when_given(api_key: Optional[str] = None) -> None:
    from evdspy.EVDSlocal.config.apikey_class import ApikeyClass

    if api_key is None:
        return
    if ApikeyClass().get_valid_api_key(check=False) is False:
        from evdspy.EVDSlocal.initial.load_commands_cmds_to_load import save_apikey

        save_apikey(api_key)


def create_metadata_url(series_code: str, response_type: str = 'json') -> str:
    base_url = "https://evds2.tcmb.gov.tr/service/evds/serieList"
    return f"{base_url}?type={response_type}&code={series_code}&mode=2"


def get_metadata_for_index(index: Union[str, tuple[Any, ...]], proxy_manager: Any, cache=True) -> pd.DataFrame:
    from evdspy.EVDSlocal.index_requests.user_requests import ApiRequester, DataProcessor
    from evdspy.EVDSlocal.index_requests.user_requests.Url_builder import HelpUrlBuilder

    config = RequestConfig(index, cache=False)  # TODO

    def get(index_local):
        config_local = RequestConfig(index_local, cache=cache)
        help_url_builder = HelpUrlBuilder(config_local)
        api_requester = ApiRequester(help_url_builder, proxy_manager)
        metadata_processor = DataProcessor(api_requester())
        return metadata_processor()

    metas = tuple(map(get, config.index))
    return pd.concat(metas, axis="rows", join="outer")


@dataclass
class Result:
    """
    get_series_exp function returns instance of this class
    data : pd.DataFrame
    metadata : pd.DataFrame
    write : creates and Excel file with data and metadadata of two sheets
    to_excel : same with write to meet pandas to_excel function
    """
    data: field(default_factory=pd.DataFrame)
    metadata: field(default_factory=pd.DataFrame)
    write: Callable

    def __post_init__(self):
        self.to_excel = self.write

    def __str__(self):
        content = super.__str__(self)

        return """
    ! get_series_exp function returns instance of this class 
    <Result>
    data     : pd.DataFrame  => contains data [ same as what get_series function returns ] 
    metadata : pd.DataFrame  => contains metadata if available 
    write    : Callable      => creates and Excel file with data and metadadata of two sheets 
    to_excel : Callable      =>  same with write to meet pandas to_excel function 
    
        """ + content


def get_series_exp(
        index: Union[str, tuple[Any, ...]],
        start_date: str = default_start_date_fnc(),
        end_date: str = default_end_date_fnc(),
        frequency: Union[
            Literal[
                "monthly", "weekly", "annually", "semimonthly", "semiannually", "business", None
            ]
        ] = None,
        formulas: Union[Literal["level", "percentage_change", "difference"], None] = None,
        aggregation: Union[
            Literal["avg", "min", "max", "first", "last", "sum", None], None
        ] = None,
        cache: bool = False ,
        meta_cache: bool = False ,
        proxy: Optional[str] = None,
        proxies: Optional[dict[str, str]] = None,
        debug: bool = False,
        api_key: Optional[str] = None,
) -> dict[str, pd.DataFrame]:
    """
    Retrieves economic data series from the specified API and returns it as a pandas DataFrame.
    Parameters
    ----------
    index : str or tuple of str
        The identifier(s) for the data series to fetch. Can be a single string for one series or a tuple of strings for multiple series.
    start_date : str, optional
        The start date for the data retrieval in 'DD-MM-YYYY' format, by default calls default_start_date_fnc().
    end_date : str, optional
        The end date for the data retrieval in 'DD-MM-YYYY' format, by default calls default_end_date_fnc().
    frequency : str, optional
        The frequency at which data should be retrieved.
        monthly | weekly | annually | semimonthly | semiannually | business
    formulas : str or tuple of str, optional
        The computation methods to apply to the data series
        level | percentage_change | difference | year_to_year_percent_change | year_to_year_differences
    aggregation : str or tuple of str, optional
        The aggregation methods to apply to the data, similar to formulas.
        avg |min | max | first | last | sum
    cache : bool, optional
        If True, uses cached data when available to speed up the data retrieval process, by default False.
    meta_cache : bool, optional
        If True, uses cached data for metadata when available to speed up the data retrieval process, by default False.
    proxy : str, optional
        The URL of the proxy server to use for the requests, by default None.
    proxies : dict, optional
        A dictionary of proxies to use for the request, by default None.
    debug : bool, optional
        If True, runs the function in debug mode, providing additional debug information without making a real API request, by default False.
    api_key : str, optional
        The API key required for accessing the data, by default None.
        When it was given for the first time it will be saved to a file for the subsequent requests.
        alternatively it may be saved by save("APIKEY") function or $ evdspy save [from console]
    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing the retrieved data series.
    Raises
    ------
    ValueError
        If an invalid API key is provided or required parameters are missing.
    """

    from evdspy.EVDSlocal.index_requests.user_requests import (
        RequestConfig,
        ProxyManager,
        UrlBuilder,
        DataProcessor,
    )
    from evdspy.EVDSlocal.index_requests.user_requests.Api_requester import ApiRequester

    # ............initial_api_process_when_given...............
    initial_api_process_when_given(api_key)
    # ............RequestConfig................................
    config = RequestConfig(
        index=index,
        start_date=start_date,
        end_date=end_date,
        frequency=frequency,
        formulas=formulas,
        aggregation=aggregation,
        cache=cache,
    )
    # # ............ProxyManager................................
    proxy_manager = ProxyManager(proxy=proxy, proxies=proxies)
    # # ............UrlBuilder..................................
    # url_builder = UrlBuilder(config, url_type=None)
    # # ............ApiRequester................................
    # api_requester = ApiRequester(url_builder, proxy_manager)
    # if debug:
    #     return api_requester.dry_request()
    # # ............DataProcessor................................
    # data_processor = DataProcessor(api_requester())

    # main_data = data_processor()
    main_data = get_series(index ,
                           start_date=start_date,
        end_date=end_date,
        frequency=frequency,
        formulas=formulas,
        aggregation=aggregation,
        cache=cache )

    metadata_: pd.DataFrame = get_metadata_for_index(index, proxy_manager, cache=meta_cache)

    result = {
        "main_data": main_data,
        "metadata": metadata_,
    }

    def build_df(file_name='output.xlsx'):
        res = False
        with pd.ExcelWriter(file_name) as writer:
            for sheet_name, df in result.items():
                if isinstance(df, pd.DataFrame):
                    df.to_excel(writer, sheet_name=sheet_name, index=False)
                    res = True
        if res:
            print(f" writing file :  [{file_name}] ")

    return Result(main_data, metadata_, build_df)


def test_get_series2(capsys):
    with capsys.disabled():
        # setup()
        result = get_series_exp("TP.ODEMGZS.BDTTOPLAM", cache=False)
        assert isinstance(result['main_data'], pd.DataFrame)
        assert isinstance(result['metadata'], pd.DataFrame)


__all__ = ("get_series_exp",)
